import logging
import numpy as np


logger = logging.getLogger(__name__)

__all__ = ["purcell", "sinusoidal", "general_toroidal"]


def general_toroidal(
    phase,
    profile_equation='elliptical',
    mode='max',
    target_Pc=None,
    num_points=1000,
    throat_scale_a='throat.scale_a',
    throat_scale_b='throat.scale_b',
    throat_diameter='throat.diameter',
    touch_length='throat.touch_length',
    surface_tension='throat.surface_tension',
    contact_angle='throat.contact_angle'
):
    r"""
    The general model for meniscus properties inside a toroidal throat

    Parameters
    ----------
    %(phase)s
    profile_equation : str
        Options are 'elliptical' (default), 'sinusoidal'.
    mode : str
        Determines what information to send back. Options are:

            ======== =========================================================
            Mode     Description
            ======== =========================================================
            'max'    (default) The maximum capillary pressure along the throat
                     axis
            'touch'  The maximum capillary pressure a meniscus can sustain
                     before touching a solid feature
            ======== =========================================================

    target_Pc : float
        The target capillary pressure to return data for when mode is 'men'
    num_points : float
        The number of divisions to make along the profile length to assess the
        meniscus properties in order to find target pressures, touch lengths,
        minima and maxima. The default is 1000.
    throat_scale_a : str
        %(dict_blurb)s scale factor for adjusting the profile along the
        throat axis (x).
    throat_scale_b : str
        %(dict_blurb)s scale factor for adjusting the profile perpendicular
        to the throat axis (y).
    throat_diameter : str
        %(dict_blurb)s throat diameter values to be used.
    touch_length : str
        %(dict_blurb)s maximum length that a meniscus can
        protrude into the connecting pore before touching a solid feature and
        therfore invading
    surface_tension : str
        %(dict_blurb)s surface tension values to be used. If a pore property
        is given, it is interpolated to a throat list.
    contact_angle : str
        %(dict_blurb)s contact angle values to be used. If a pore property
        is given, it is interpolated to a throat list.

    Returns
    -------
    values : dict
        A dictionary containing the following entries:

        ============ =========================================================
        Item         Description
        ============ =========================================================
        'pos'        xpos
        'rx'         rx(xpos, fa, fb, throatRad)
        'alpha'      fill_angle(xpos, fa, fb, throatRad)
        'alpha_min'  fill_angle(xmin, fa, fb, throatRad)
        'alpha_max'  fill_angle(xmax, fa, fb, throatRad)
        'c2x'        c2x(xpos, fa, fb, throatRad, contact)
        'gamma'      cap_angle(xpos, fa, fb, throatRad, contact)
        'radius'     rad_curve(xpos, fa, fb, throatRad, contact)
        'center'     (xpos - men_data['c2x'])
        'men_max'    ??
        ============ =========================================================

    """
    from sympy import symbols, lambdify
    from sympy import atan as sym_atan
    from sympy import cos as sym_cos
    from sympy import sin as sym_sin
    from sympy import sqrt as sym_sqrt
    from sympy import pi as sym_pi

    # Get data from dictionary keys
    network = phase.network
    surface_tension = phase[surface_tension]
    contact = phase[contact_angle]
    # Contact Angle in radians
    contact = np.deg2rad(contact)
    # Network properties
    throatRad = network[throat_diameter]/2
    # Scaling parameters for throat profile
    fa = phase[throat_scale_a]
    fb = phase[throat_scale_b]
    # Governing equations
    x, a, b, rt, sigma, theta = symbols('x, a, b, rt, sigma, theta')
    if profile_equation == 'elliptical':
        y = sym_sqrt(1 - (x/a)**2)*b
    elif profile_equation == 'sinusoidal':
        y = (sym_cos((sym_pi/2)*(x/a)))*b
    else:
        logger.error('Profile equation is not valid, default to elliptical')
        y = sym_sqrt(1 - (x/a)**2)*b
    # Throat radius profile
    r = rt + (b-y)
    # Derivative of profile
    rprime = r.diff(x)
    # Filling angle
    alpha = sym_atan(rprime)
    # Angle between y axis and contact point to meniscus center
    eta = sym_pi - alpha - theta
    gamma = sym_pi/2 - eta
    # Radius of curvature of meniscus
    rm = r/sym_cos(eta)
    # distance from center of curvature to meniscus contact point (Pythagoras)
    d = rm*sym_sin(eta)
    # angle between throat axis, meniscus center and meniscus contact point
    # Capillary Pressure
    p = 2*sigma/rm
    # Callable functions
    rx = lambdify((x, a, b, rt), r, 'numpy')
    fill_angle = lambdify((x, a, b, rt), alpha, 'numpy')
    rad_curve = lambdify((x, a, b, rt, theta), rm, 'numpy')
    c2x = lambdify((x, a, b, rt, theta), d, 'numpy')
    cap_angle = lambdify((x, a, b, rt, theta), gamma, 'numpy')
    Pc = lambdify((x, a, b, rt, theta, sigma), p, 'numpy')
    # All relative positions along throat
    hp = int(num_points/2)
    log_pos = np.logspace(-4, -1, hp+1)[:-1]
    lin_pos = np.arange(0.1, 1.0, 1/hp)
    half_pos = np.concatenate((log_pos, lin_pos))
    pos = np.concatenate((-half_pos[::-1], half_pos))
    # Now find the positions of the menisci along each throat axis
    Y, X = np.meshgrid(throatRad, pos)
    X *= fa
    # throat Capillary Pressure
    t_Pc = Pc(X, fa, fb, Y, contact, surface_tension)
    # Values of minima and maxima
    Pc_min = np.min(t_Pc, axis=0)
    Pc_max = np.max(t_Pc, axis=0)
    # Arguments of minima and maxima
    a_min = np.argmin(t_Pc, axis=0)
    a_max = np.argmax(t_Pc, axis=0)
    if mode == 'max':
        return Pc_max
    elif mode == 'touch':
        all_rad = rad_curve(X, fa, fb, Y, contact)
        all_c2x = c2x(X, fa, fb, Y, contact)
        all_cen = X - all_c2x
        dist = all_cen + all_rad
        # Only count lengths where meniscus bulges into pore
        dist[all_rad < 0] = 0.0
        touch_len = network[touch_length]
        mask = dist > touch_len
        arg_touch = np.argmax(mask, axis=0)
        # Make sure we only count ones that happen before max pressure
        # And above min pressure (which will be erroneous)
        arg_in_range = (arg_touch < a_max) * (arg_touch > a_min)
        arg_touch[~arg_in_range] = a_max[~arg_in_range]
        x_touch = pos[arg_touch]*fa
        # Return the pressure at which a touch happens
        Pc_touch = Pc(x_touch, fa, fb, throatRad, contact, surface_tension)
        return Pc_touch
    elif (mode == 'men'):
        if target_Pc is None:
            logger.error(msg='Please supply a target capillary pressure'
                         + ' when mode is "men", defaulting to 1.0e-6')
            target_Pc = 1.0e-6
    else:
        raise Exception('Unrecognized mode')
    if np.abs(target_Pc) < 1.0e-6:
        logger.error(msg='Please supply a target capillary pressure'
                     + ' with absolute value greater than 1.0e-6,'
                     + ' default to 1.0e-6')
        target_Pc = 1.0e-6
    # Find the position in-between the minima and maxima corresponding to
    # the target pressure
    inds = np.indices(np.shape(t_Pc))
    # Change values outside the range between minima and maxima to be those
    # Values
    mask = inds[0] < np.ones(len(pos))[:, np.newaxis]*a_min
    t_Pc[mask] = (np.ones(len(pos))[:, np.newaxis]*Pc_min)[mask]
    mask = inds[0] > np.ones(len(pos))[:, np.newaxis]*a_max
    t_Pc[mask] = (np.ones(len(pos))[:, np.newaxis]*Pc_max)[mask]
    # Find the argument at or above the target Pressure
    mask = t_Pc >= target_Pc
    arg_x = np.argmax(mask, axis=0)
    # If outside range change to minima or maxima accordingly
    arg_x[target_Pc < Pc_min] = a_min[target_Pc < Pc_min]
    arg_x[target_Pc > Pc_max] = a_max[target_Pc > Pc_max]
    xpos = pos[arg_x]*fa
    xmin = pos[a_min]*fa
    xmax = pos[a_max]*fa
    # Output
    men_data = {}
    men_data['pos'] = xpos
    men_data['rx'] = rx(xpos, fa, fb, throatRad)
    men_data['alpha'] = fill_angle(xpos, fa, fb, throatRad)
    men_data['alpha_min'] = fill_angle(xmin, fa, fb, throatRad)
    men_data['alpha_max'] = fill_angle(xmax, fa, fb, throatRad)
    men_data['c2x'] = c2x(xpos, fa, fb, throatRad, contact)
    men_data['gamma'] = cap_angle(xpos, fa, fb, throatRad, contact)
    men_data['radius'] = rad_curve(xpos, fa, fb, throatRad, contact)
    # xpos is relative to the throat center
    men_data['center'] = (xpos - men_data['c2x'])
    men_data['men_max'] = men_data['center'] - men_data['radius']

    logger.info(mode+' calculated for Pc: '+str(target_Pc))
    return men_data


def sinusoidal(
    phase,
    mode='max',
    target_Pc=None,
    num_points=1e3,
    r_toroid=5e-6,
    throat_diameter='throat.diameter',
    pore_diameter='pore.diameter',
    touch_length='throat.touch_length',
    surface_tension='throat.surface_tension',
    contact_angle='throat.contact_angle'
):
    r"""
    Wrapper for the toroidal model to implement a sinusoidal profile.

    The quarter-wavelength is equal to toroidal radius r_toroid.
    The amplitude is equal to the toroidal radius multiplied by the ratio
    of the throat radius and average connecting pore radius.

    Notes
    -----
    The capillary pressure equation for a sinusoidal throat is extended
    from the Washburn equation as [1]_:

    .. math::

        P_c = -\frac{2\sigma(cos(\alpha + \theta))}{r(x)}

    where alpha is:

    .. math::

        \alpha = arctan(\frac{dr}{dx})

    References
    ----------
    .. [1] A. Forner-Cuenca et. al, Advanced Water Management in PEFCs:
        Diffusion Layers with Patterned Wettability.
        J. ECS. 163, 9, F1038-F1048 (2016).

    """
    phase['throat.scale_a'] = r_toroid
    phase['throat.scale_b'] = r_toroid
    output = general_toroidal(phase=phase,
                              mode=mode,
                              profile_equation='sinusoidal',
                              target_Pc=target_Pc,
                              num_points=num_points,
                              throat_diameter=throat_diameter,
                              touch_length=touch_length,
                              surface_tension=surface_tension,
                              contact_angle=contact_angle)
    return output


def purcell(
    phase,
    mode='max',
    target_Pc=None,
    num_points=1e3,
    r_toroid=5e-6,
    throat_diameter='throat.diameter',
    touch_length='throat.touch_length',
    surface_tension='throat.surface_tension',
    contact_angle='throat.contact_angle'
):
    r"""
    Wrapper for the general toroidal model to implement the Purcell
    equation for a torus with cylindrical profile and toroidal radius
    r_toroid.

    Notes
    -----
    This approach accounts for the converging-diverging nature of many
    throat types. Advancing the meniscus beyond the apex of the toroid
    requires an increase in capillary pressure beyond that for a
    cylindical tube of the same radius. The details of this equation are
    described by Mason and Morrow [1], and explored by Gostick [2] in the
    context of a pore network model.

    References
    ----------
    [1] Missing reference
    [2] Missing reference

    """
    phase['throat.scale_a'] = r_toroid
    phase['throat.scale_b'] = r_toroid
    output = general_toroidal(phase=phase,
                              mode=mode,
                              profile_equation='elliptical',
                              target_Pc=target_Pc,
                              num_points=num_points,
                              throat_diameter=throat_diameter,
                              touch_length=touch_length,
                              surface_tension=surface_tension,
                              contact_angle=contact_angle)
    return output



import logging
import numpy as np
import pandas as pd
from xml.etree import ElementTree as ET
from openpnm.io import project_to_dict, _parse_filename
from openpnm.utils import Workspace
from openpnm.utils._misc import is_transient
logger = logging.getLogger(__name__)
ws = Workspace()


_TEMPLATE = """
<?xml version="1.0" ?>
<VTKFile byte_order="LittleEndian" type="PolyData" version="0.1">
    <PolyData>
        <Piece NumberOfLines="0" NumberOfPoints="0">
            <Points>
            </Points>
            <Lines>
            </Lines>
            <PointData>
            </PointData>
            <CellData>
            </CellData>
        </Piece>
    </PolyData>
</VTKFile>
""".strip()


def project_to_vtk(project, filename="",
                   fill_nans=None, fill_infs=None):
    r"""
    Save network and phase data to a single vtp file for visualizing in
    Paraview.

    Parameters
    ----------
    network : Network
        The Network containing the data to be written
    phases : list, optional
        A list containing Phase(s) containing data to be
        written
    filename : str, optional
        Filename to write data.  If no name is given the file is named
        after the network
    fill_nans : scalar
        The value to use to replace NaNs with.  The VTK file format does
        not work with NaNs, so they must be dealt with.  The default is
        `None` which means property arrays with NaNs are not written to the
        file.  Other useful options might be 0 or -1, but the user must
        be aware that these are not real values, only place holders.
    fill_infs : scalar
        The value to use to replace infs with.  The default is ``None``
        which means that property arrays containing ``None`` will *not*
        be written to the file, and a warning will be issued.  A useful
        value is

    """
    network = project.network
    algs = project.algorithms
    # Check if any of the phases has time series
    transient = is_transient(algs)
    if transient:
        logger.warning(
            "vtp format does not support transient data, " + "use xdmf instead"
        )
    if filename == "":
        filename = project.name
    filename = _parse_filename(filename=filename, ext="vtp")

    am = project_to_dict(project=project,
                         categorize_by=["object", "data"])
    am = pd.json_normalize(am, sep='.').to_dict(orient='records')[0]
    for k in list(am.keys()):
        am[k.replace('.', ' | ')] = am.pop(k)
    key_list = list(sorted(am.keys()))

    points = network["pore.coords"]
    pairs = network["throat.conns"]
    num_points = np.shape(points)[0]
    num_throats = np.shape(pairs)[0]

    root = ET.fromstring(_TEMPLATE)
    piece_node = root.find("PolyData").find("Piece")
    piece_node.set("NumberOfPoints", str(num_points))
    piece_node.set("NumberOfLines", str(num_throats))
    points_node = piece_node.find("Points")
    coords = _array_to_element("coords", points.T.ravel("F"), n=3)
    points_node.append(coords)
    lines_node = piece_node.find("Lines")
    connectivity = _array_to_element("connectivity", pairs)
    lines_node.append(connectivity)
    offsets = _array_to_element("offsets", 2 * np.arange(len(pairs)) + 2)
    lines_node.append(offsets)

    point_data_node = piece_node.find("PointData")
    cell_data_node = piece_node.find("CellData")
    for key in key_list:
        array = am[key]
        if array.dtype == "O":
            logger.warning(key + " has dtype object," + " will not write to file")
        else:
            if array.dtype == bool:
                array = array.astype(int)
            if np.any(np.isnan(array)):
                if fill_nans is None:
                    logger.warning(key + " has nans," + " will not write to file")
                    continue
                else:
                    array[np.isnan(array)] = fill_nans
            if np.any(np.isinf(array)):
                if fill_infs is None:
                    logger.warning(key + " has infs," + " will not write to file")
                    continue
                else:
                    array[np.isinf(array)] = fill_infs
            element = _array_to_element(key, array)
            if array.size == num_points:
                point_data_node.append(element)
            elif array.size == num_throats:
                cell_data_node.append(element)

    tree = ET.ElementTree(root)
    tree.write(filename)

    with open(filename, "r+") as f:
        string = f.read()
        string = string.replace("</DataArray>", "</DataArray>\n\t\t\t")
        f.seek(0)
        # consider adding header: '<?xml version="1.0"?>\n'+
        f.write(string)


def _array_to_element(name, array, n=1):
    dtype_map = {
        "int8": "Int8",
        "int16": "Int16",
        "int32": "Int32",
        "int64": "Int64",
        "uint8": "UInt8",
        "uint16": "UInt16",
        "uint32": "UInt32",
        "uint64": "UInt64",
        "float32": "Float32",
        "float64": "Float64",
        "str": "String",
    }
    element = None
    if str(array.dtype) in dtype_map.keys():
        element = ET.Element("DataArray")
        element.set("Name", name)
        element.set("NumberOfComponents", str(n))
        element.set("type", dtype_map[str(array.dtype)])
        element.text = "\t".join(map(str, array.ravel()))
    return element


def _element_to_array(element, n=1):
    string = element.text
    dtype = element.get("type")
    array = np.fromstring(string, sep="\t")
    array = array.astype(dtype.lower())
    if n != 1:
        array = array.reshape(array.size // n, n)
    return array



from scipy.integrate import solve_ivp

from openpnm.algorithms._solution import TransientSolution
from openpnm.integrators import Integrator

__all__ = ['ScipyRK45']


class ScipyRK45(Integrator):
    """Integrator class based on SciPy's implementation of RK45"""

    def __init__(self, atol=1e-6, rtol=1e-6, verbose=False, linsolver=None):
        self.atol = atol
        self.rtol = rtol
        self.verbose = verbose
        self.linsolver = linsolver

    def solve(self, rhs, x0, tspan, saveat, **kwargs):
        """
        Solves the system of ODEs defined by dy/dt = rhs(t, y).

        Parameters
        ----------
        rhs : function handle
            RHS vector in the system of ODEs defined by dy/dt = rhs(t, y)
        x0 : array_like
            Initial value for the system of ODEs
        tspan : array_like
            2-element tuple (or array) representing the timespan for the
            system of ODEs
        saveat : float or array_like
            If float, defines the time interval at which the solution is
            to be stored. If array_like, defines the time points at which
            the solution is to be stored.
        **kwargs : keyword arguments
            Other keyword arguments that might get used by the integrator

        Returns
        -------
        TransientSolution
            Solution of the system of ODEs stored in a subclass of numpy's
            ndarray with some added functionalities (ex. you can get the
            solution at intermediate time points via: y = soln(t_i)).

        """
        options = {
            "atol": self.atol,
            "rtol": self.rtol,
            "t_eval": saveat,
            # FIXME: uncomment next line when/if scipy#11815 is merged
            # "verbose": self.verbose,
        }
        sol = solve_ivp(rhs, tspan, x0, method="RK45", **options)
        if sol.success:
            return TransientSolution(sol.t, sol.y)
        raise Exception(sol.message)



import openpnm.models as mods


standard_gas = {
    'pore.density': {
        'model': mods.phase.density.ideal_gas,
    },
    'pore.heat_capacity_gas': {
        'model': mods.phase.heat_capacity.gas_pure_TRC,
    },
    'pore.heat_capacity': {
        'model': mods.phase.heat_capacity.gas_pure_TRC,
    },
    'pore.thermal_conductivity': {
        'model': mods.phase.thermal_conductivity.gas_pure_gismr,
    },
    'pore.viscosity': {
        'model': mods.phase.viscosity.gas_pure_gesmr,
    },
}


standard_gas_mixture = {
    'pore.density': {
        'model': mods.phase.density.ideal_gas,
    },
    'pore.heat_capacity': {
        'model': mods.phase.heat_capacity.gas_mixture_yweighted,
    },
    'pore.thermal_conductivity': {
        'model': mods.phase.thermal_conductivity.gas_mixture_whz,
    },
    'pore.viscosity': {
        'model': mods.phase.viscosity.gas_mixture_hz,
    },
}

binary_gas_mixture = {
    'pore.density': {
        'model': mods.phase.density.ideal_gas,
    },
    'pore.heat_capacity': {
        'model': mods.phase.heat_capacity.gas_mixture_yweighted,
    },
    'pore.thermal_conductivity': {
        'model': mods.phase.thermal_conductivity.gas_mixture_whz,
    },
    'pore.viscosity': {
        'model': mods.phase.viscosity.gas_mixture_hz,
    },
    'pore.diffusivity': {
        'model': mods.phase.diffusivity.gas_mixture_ce,
    },
}



import numpy as _np
from openpnm.models import misc as _misc
from openpnm.models.geometry import _geodocs


__all__ = ["weibull",
           "normal",
           "generic_distribution",
           "random",
           "from_neighbor_pores",
           "equivalent_diameter"]


def weibull(network, shape, scale, loc, seeds='throat.seed'):
    return _misc.weibull(network, shape=shape, scale=scale, loc=loc,
                         seeds=seeds)


weibull.__doc__ = _misc.weibull.__doc__


def normal(network, scale, loc, seeds='throat.seed'):
    return _misc.normal(network, scale=scale, loc=loc, seeds=seeds)


normal.__doc__ = _misc.normal.__doc__


def generic_distribution(network, func, seeds='throat.seed'):
    return _misc.generic_distribution(network, func=func, seeds=seeds)


generic_distribution.__doc__ = _misc.generic_distribution.__doc__


def random(network, seed=None, num_range=[0, 1]):
    return _misc.random(network, element='throat', seed=seed,
                        num_range=num_range)


random.__doc__ = _misc.random.__doc__


def from_neighbor_pores(network, prop='pore.diameter', mode='min'):
    return _misc.from_neighbor_pores(network, prop=prop,
                                     mode=mode)


from_neighbor_pores.__doc__ = _misc.from_neighbor_pores.__doc__


@_geodocs
def equivalent_diameter(
    network,
    throat_area='throat.cross_sectional_area',
    throat_shape='circle',
):
    r"""
    Calculates the diameter of a cirlce or edge-length of a sqaure with same
    area as the throat.

    Parameters
    ----------
    %(network)s
    %(At)s
    throat_shape : str
        The shape cross-sectional shape of the throat to assume when
        back-calculating from the area.  Options are 'circle' (default) or
        'square'.

    Returns
    -------
    diameters : ndarray
        A numpy ndarray containing throat diameter values

    """
    area = network[throat_area]
    if throat_shape.startswith('circ'):
        value = 2*_np.sqrt(area/_np.pi)
    elif throat_shape.startswith('square'):
        value = _np.sqrt(area)
    return value



import logging
from openpnm.algorithms import TransientReactiveTransport, AdvectionDiffusion
from openpnm.utils import Docorator


__all__ = ['TransientAdvectionDiffusion']


docstr = Docorator()
logger = logging.getLogger(__name__)


@docstr.dedent
class TransientAdvectionDiffusionSettings:
    r"""
    Parameters
    ----------
    %(AdvectionDiffusionSettings.parameters)s
    %(TransientReactiveTransportSettings.parameters)s

    """


class TransientAdvectionDiffusion(TransientReactiveTransport,
                                  AdvectionDiffusion):
    r"""
    A subclass of Transport to perform steady and transient simulations
    of pure diffusion and advection-diffusion problems.
    """

    def __init__(self, name='trans_ad_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(TransientAdvectionDiffusionSettings())



r"""
Pore-scale models for calculating the conductance of conduits.
"""
from numpy import vstack


__all__ = [
    '_poisson_conductance',
    '_get_key_props',
]


def _poisson_conductance(phase,
                         pore_conductivity=None,
                         throat_conductivity=None,
                         size_factors=None):
    r"""
    Calculates the conductance of the conduits in the network, where a
    conduit is (1/2 pore - full throat - 1/2 pore). See the notes section.

    Parameters
    ----------
    phase : OpenPNM Phase
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    pore_conductivity : str
        Dictionary key of the pore conductivity values
    throat_conductivity : str
        Dictionary key of the throat conductivity values
    size_factors: str
        Dictionary key of the conduit size factors' values.

    Returns
    -------
    g : ndarray
        Array containing conductance values for conduits in the
        geometry attached to the given physics object.

    Notes
    -----
    This function requires that all the necessary phase properties
    already be calculated.

    """
    network = phase.network
    cn = network.conns
    Dt = phase[throat_conductivity]
    D1, D2 = phase[pore_conductivity][cn].T
    # If individual size factors for conduit constiuents are known
    SF = network[size_factors]
    if SF.ndim == 2:
        F1, Ft, F2 = SF.T
        g1 = D1 * F1
        gt = Dt * Ft
        g2 = D2 * F2
        return 1 / (1 / g1 + 1 / gt + 1 / g2)
    else:
        # Otherwise, i.e., the size factor for the entire conduit is only known
        F = network[size_factors]
        return Dt * F


def _get_key_props(phase=None,
                   diameter="throat.diameter",
                   surface_tension="throat.surface_tension",
                   contact_angle="throat.contact_angle"):
    """
    Returns desired properties in the correct format! See Notes.

    Notes
    -----
    Many of the methods are generic to pores and throats. Some information may
    be stored on either the pore or throat and needs to be interpolated.
    This is a helper method to return the properties in the correct format.

    TODO: Check for method to convert throat to pore data

    """
    element = diameter.split(".")[0]
    if element == "pore":
        if "throat" in surface_tension:
            sigma = phase.interpolate_data(propname=surface_tension)
        else:
            sigma = phase[surface_tension]
        if "throat" in contact_angle:
            theta = phase.interpolate_data(propname=contact_angle)
        else:
            theta = phase[contact_angle]
    if element == "throat":
        if "pore" in surface_tension:
            sigma = phase.interpolate_data(propname=surface_tension)
        else:
            sigma = phase[surface_tension]
        if "pore" in contact_angle:
            theta = phase.interpolate_data(propname=contact_angle)
        else:
            theta = phase[contact_angle]

    return element, sigma, theta



from collections import namedtuple

import numpy as np
import scipy.sparse as sprs
from scipy.sparse import csgraph

from openpnm._skgraph.operations import split_edges

__all__ = [
    'trim_disconnected_clusters',
    'ispercolating',
    'remove_isolated_clusters',
    'bond_percolation',
    'site_percolation',
    # 'find_connected_clusters',
    # 'find_trapped_bonds',
    # 'find_trapped_sites',
    'find_connected_clusters',
]


def bond_percolation(conns, occupied_bonds):
    r"""
    Assigns cluster numbers to sites and bonds acccording to a bond
    percolation process, given a list of occupied bonds.

    Parameters
    ----------
    conns : array_like
        An N x 2 array connections. Any sites connected to an occupied bond
        will also be considered occupied and given the same cluster number
        as the bond.
    occupied_bonds : ndarray
        A boolean array with one element for each bond, with ``True`` values
        indicating that a bond is occupied

    Returns
    -------
    A tuple containing a list of site and bond labels, indicating which
    cluster each belongs to. A value of -1 indicates uninvaded.

    Notes
    -----
    The ``connected_components`` function of ``scipy.sparse.csgraph`` will give
    a cluster number to ALL bonds whether they are occupied or not, so this
    function essentially adjusts the cluster numbers to represent a
    percolation process.

    """
    Np = np.amax(conns) + 1
    # Find occupied sites based on status of shared bonds
    occupied_sites = np.zeros([Np, ], dtype=bool)
    occupied_sites[conns[occupied_bonds].flatten()] = True
    # Perform cluster labeling of network
    adj_mat = sprs.csr_matrix((occupied_bonds.astype(int),
                               (conns[:, 0], conns[:, 1])),
                              shape=(Np, Np))
    adj_mat.eliminate_zeros()
    clusters = csgraph.connected_components(csgraph=adj_mat, directed=False)[1]
    # Set cluster number of unoccupied sites to -1
    s_labels = (clusters + 1)*occupied_sites - 1
    # Bonds inherit the cluster number of its connected sites
    b_labels = np.amin(s_labels[conns], axis=1)
    # Set cluster number of unoccupied bonds to -1
    b_labels[~occupied_bonds] = -1
    tup = namedtuple('cluster_labels', ('site_labels', 'bond_labels'))
    return tup(s_labels, b_labels)


def site_percolation(conns, occupied_sites):
    r"""
    Assigns cluster numbers to sites and bonds acccording to a site
    percolation process, given a list of occupied sites.

    Parameters
    ----------
    conns : array_like
        An N x 2 array connections. If two connected sites are both occupied
        they are part of the same cluster, as is the bond connecting them.
    occupied_sites : ndarray
        A boolean array with one element for each site, with ``True`` values
        indicating that a site is occupied

    Returns
    -------
    A tuple containing a list of site and bond labels, indicating which
    cluster each belongs to.  A value of -1 indicates unoccupied.

    Notes
    -----
    The ``connected_components`` function of ``scipy.sparse.csgraph`` will
    give ALL sites a cluster number whether they are occupied or not, so this
    function essentially adjusts the cluster numbers to represent a
    percolation process.

    """
    Np = np.size(occupied_sites)
    # Find bond occupancy based on status of its connected sites
    occupied_bonds = np.all(occupied_sites[conns], axis=1)
    # Perform cluster labeling of network
    adj_mat = sprs.csr_matrix((occupied_bonds, (conns[:, 0], conns[:, 1])),
                              shape=(Np, Np))
    adj_mat.eliminate_zeros()
    clusters = csgraph.connected_components(csgraph=adj_mat, directed=False)[1]
    # Set cluster number of unoccupied sites to -1
    s_labels = (clusters + 1)*occupied_sites - 1
    # Bonds inherit the cluster number of its connected sites
    b_labels = np.amin(s_labels[conns], axis=1)
    # Set cluster number of unoccupied bonds to -1
    b_labels[~occupied_bonds] = -1
    tup = namedtuple('cluster_labels', ('site_labels', 'bond_labels'))
    return tup(s_labels, b_labels)


def mixed_percolation(
    conns,
    occupied_sites,
    occupied_bonds
):  # pragma: no cover
    new_conns = split_edges(conns)[0]
    new_sites = np.hstack((occupied_sites, occupied_bonds))
    s, b = site_percolation(conns=new_conns, occupied_sites=new_sites)
    s_labels = s[:occupied_sites.shape[0]]
    b_labels = s[occupied_sites.shape[0]:]
    return s_labels, b_labels


def find_connected_clusters(
    bond_labels,
    site_labels,
    inlets,
    asmask=True
):  # pragma: no cover
    hits = np.unique(site_labels[inlets])
    hits = hits[hits >= 0]
    occupied_bonds = np.isin(bond_labels, hits)
    occupied_sites = np.isin(site_labels, hits)
    if not asmask:
        occupied_bonds = occupied_bonds*(bond_labels + 1) - 1
        occupied_sites = occupied_sites*(site_labels + 1) - 1
    return occupied_sites, occupied_bonds


def find_trapped_bonds(conns, outlets, occupied_bonds):  # pragma: no cover
    s_labels, b_labels = bond_percolation(conns, ~occupied_bonds)
    s_labels2, b_labels2 = find_connected_clusters(b_labels, s_labels,
                                                   outlets, asmask=False)
    # Set sites and bonds connected to outlets to -1, keeping
    s_labels[s_labels2 >= 0] = -1
    b_labels[b_labels2 >= 0] = -1
    return s_labels, b_labels


def find_trapped_sites(conns, outlets, occupied_sites):  # pragma: no cover
    s_labels, b_labels = site_percolation(conns, ~occupied_sites)
    s_labels2, b_labels2 = find_connected_clusters(b_labels, s_labels,
                                                   outlets, asmask=False)
    # Set sites and bonds connected to outlets to -1, keeping
    s_labels[s_labels2 >= 0] = -1
    b_labels[b_labels2 >= 0] = -1
    return s_labels, b_labels


def trim_disconnected_clusters(b_labels, s_labels, inlets):
    r"""
    Computes actual node and edge occupancy based on connectivity to the given
    inlets

    Parameters
    ----------
    b_labels : ndarray
        An array of cluster labels assigned to each bond.  -1 indicates
        unoccupied
    s_labels : ndarray
        An array of cluster labels assigned to each site. -1 indicates
        unoccupied. Site cluster numbers must correspond to the bond
        clusters, such that if bond j has a cluster number N, then both
        sites on each end of j are also labeled N.
    inlets : ndarray
        An array containing node indices that are to be treated as inlets.
        Any clusters labels not found in these nodes will be considered
        disconnected and set to -1.

    Returns
    -------
    occupancy : tuple of ndarrays
        The returned tuple containing arrays of cluster numbers of
        ``occupied_sites`` and ``occupied_bonds``, after accounting for
        connection to the ``inlets``.

    Notes
    -----
    The ``b_labels`` and ``s_labels`` arrays are returned from the
    ``bond_percolation`` or ``site_percolation`` function.

    """
    hits = np.unique(s_labels[inlets])
    hits = hits[hits >= 0]
    occupied_bonds = np.isin(b_labels, hits)*(b_labels + 1) - 1
    occupied_sites = np.isin(s_labels, hits)*(s_labels + 1) - 1
    r = namedtuple('cluster_labels', ('site_labels', 'bond_labels'))
    return r(occupied_sites, occupied_bonds)


def remove_isolated_clusters(labels, inlets):
    r"""
    Finds cluster labels not attached to the inlets, and sets them to
    unoccupied (-1)

    Parameters
    ----------
    labels : tuple of node and edge labels
        This information is provided by the ``site_percolation`` or
        ``bond_percolation`` functions
    inlets : array_like
        A list of which nodes are inlets.  Can be a boolean mask or an
        array of indices.

    Returns
    -------
    A tuple containing a list of node and edge labels, with all clusters
    not connected to the inlets set to not occupied (-1).

    """
    # Identify clusters of invasion sites
    inv_clusters = np.unique(labels.site_labels[inlets])
    # Remove cluster numbers == -1, if any
    inv_clusters = inv_clusters[inv_clusters >= 0]
    # Find all pores in invading clusters
    p_invading = np.in1d(labels.site_labels, inv_clusters)
    labels.site_labels[~p_invading] = -1
    t_invading = np.in1d(labels.bond_labels, inv_clusters)
    labels.bond_labels[~t_invading] = -1
    return labels


def ispercolating(conns, occupied, inlets, outlets):
    r"""
    Determines if a percolating cluster exists in the network spanning
    the given inlet and outlet nodes

    Parameters
    ----------
    conns : array_like
        An N x 2 array connections. If two connected sites are both occupied
        they are part of the same cluster, as is the bond connecting them.
    occupied : array_like
        A boolean array with ``True`` values indicating if a bond or site
        is occupied.  If the length of this array is equal to the number
        of bonds (i.e. ``conns.shape[0]``) then bond percolation is assumed,
        otherwise site percolation is assumed.
    inlets : array_like
        An array of indices indicating which nodes are part of the inlets
    outlets : array_like
        An array of indices indicating which nodes are part of the outlets
    """
    if occupied.size == conns.shape[0]:
        mode = 'bond'
    else:
        mode = 'site'
    if mode.startswith('site'):
        clusters = site_percolation(conns, occupied)
    elif mode.startswith('bond'):
        clusters = bond_percolation(conns, occupied)
    ins = np.unique(clusters.site_labels[inlets])
    if ins[0] == -1:
        ins = ins[1:]
    outs = np.unique(clusters.site_labels[outlets])
    if outs[0] == -1:
        outs = outs[1:]
    hits = np.in1d(ins, outs)
    return np.any(hits)



r"""
Contrib
-------

This module contains contributions to OpenPNM, both from the developers as
well as users that are not ready to be officially included but are worth
sharing. These are subject to api changes, being moved, and potentially
removed.

"""

from ._multiphase import *
from ._transient_multiphysics import *



import logging
import pandas as pd
import xml.etree.cElementTree as ET
from openpnm.io import project_to_dict, _parse_filename
from openpnm.utils._misc import is_transient


logger = logging.getLogger(__name__)
_header = """<?xml version="1.0" ?>
             <!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>"""


def project_to_xdmf(project, filename=''):
    r"""
    Saves (transient/steady-state) data from the given objects into
    the specified file.

    Parameters
    ----------
    network : Network
        The network containing the desired data
    phases : list[Phase] (optional, default is None)
        A list of phase objects whose data are to be included

    Notes
    -----
    This method only saves the data, not any of the pore-scale models
    or other attributes.

    """
    import h5py

    network = project.network
    algs = project.algorithms

    # Check if any of the phases has time series
    transient = is_transient(algs)

    if filename == '':
        filename = project.name
    path = _parse_filename(filename=filename, ext='xmf')
    # Path is a pathlib object, so slice it up as needed
    fname_xdf = path.name
    d = project_to_dict(project=project, flatten=False,
                        categorize_by=['element', 'data'])
    D = pd.json_normalize(d, sep='.').to_dict(orient='records')[0]
    for k in list(D.keys()):
        D[k.replace('.', '/')] = D.pop(k)
    # Identify time steps
    t_steps = []
    if transient:
        for key in D.keys():
            if '#' in key:
                t_steps.append(key.split('#')[1])
    t_steps = list(set(t_steps))
    t_grid = create_grid(Name="TimeSeries", GridType="Collection",
                         CollectionType="Temporal")
    # If steady-state, define '0' time step
    if not transient:
        t_steps.append('0')
    # Setup xdmf file
    root = create_root('Xdmf')
    domain = create_domain()
    # Iterate over time steps present
    for i, t_step in enumerate(t_steps):
        # Define the hdf file
        if not transient:
            fname_hdf = path.stem+".hdf"
        else:
            fname_hdf = path.stem+'#'+t_step+".hdf"
        path_p = path.parent
        f = h5py.File(path_p.joinpath(fname_hdf), "w")
        # Add coordinate and connection information to top of HDF5 file
        f["coordinates"] = network["pore.coords"]
        f["connections"] = network["throat.conns"]
        # geometry coordinates
        row, col = f["coordinates"].shape
        dims = ' '.join((str(row), str(col)))
        hdf_loc = fname_hdf + ":coordinates"
        geo_data = create_data_item(value=hdf_loc, Dimensions=dims,
                                    Format='HDF', DataType="Float")
        geo = create_geometry(GeometryType="XYZ")
        geo.append(geo_data)
        # topolgy connections
        row, col = f["connections"].shape  # col first then row
        dims = ' '.join((str(row), str(col)))
        hdf_loc = fname_hdf + ":connections"
        topo_data = create_data_item(value=hdf_loc, Dimensions=dims,
                                     Format="HDF", NumberType="Int")
        topo = create_topology(TopologyType="Polyline",
                               NodesPerElement=str(2),
                               NumberOfElements=str(row))
        topo.append(topo_data)
        # Make HDF5 file with all datasets, and no groups
        for item in list(D.keys()):
            if D[item].dtype == 'O':
                logger.warning(item + ' has dtype object,'
                               + ' will not write to file')
                del D[item]
            elif 'U' in str(D[item][0].dtype):
                pass
            elif ('#' in item and t_step == item.split('#')[1]):
                f.create_dataset(name='/'+item.split('#')[0]+'#t',
                                 shape=D[item].shape,
                                 dtype=D[item].dtype,
                                 data=D[item],
                                 compression="gzip")
            elif ('#' not in item and i == 0):
                f.create_dataset(name='/'+item, shape=D[item].shape,
                                 dtype=D[item].dtype, data=D[item],
                                 compression="gzip")
        # Create a grid
        grid = create_grid(Name=t_step, GridType="Uniform")
        time = create_time(mode='Single', Value=t_step)
        grid.append(time)
        # Add pore and throat properties
        for item in list(D.keys()):
            if item not in ['coordinates', 'connections']:
                if ("#" in item and t_step == item.split("#")[1]) or (
                    "#" not in item
                ):
                    attr_type = 'Scalar'
                    shape = D[item].shape
                    dims = (' '.join([str(i) for i in shape]))
                    if '#' in item:
                        item = item.split('#')[0]+'#t'
                        hdf_loc = fname_hdf + ":" + item
                    elif ('#' not in item and i == 0):
                        hdf_loc = fname_hdf + ":" + item
                    elif ('#' not in item and i > 0):
                        hdf_loc = path.stem + '#' + t_steps[0] + ".hdf" + ":" + item
                    attr = create_data_item(value=hdf_loc,
                                            Dimensions=dims,
                                            Format='HDF',
                                            Precision='8',
                                            DataType='Float')
                    name = item.replace('/', ' | ')
                    if 'throat' in item:
                        Center = "Cell"
                    else:
                        Center = "Node"
                    el_attr = create_attribute(Name=name, Center=Center,
                                               AttributeType=attr_type)
                    el_attr.append(attr)
                    grid.append(el_attr)
                else:
                    pass
        grid.append(topo)
        grid.append(geo)
        t_grid.append(grid)
        # CLose the HDF5 file
        f.close()
    domain.append(t_grid)
    root.append(domain)
    with open(path_p.joinpath(fname_xdf), 'w') as file:
        file.write(_header)
        file.write(ET.tostring(root).decode("utf-8"))


def create_root(Name):
    return ET.Element(Name)


def create_domain():
    return ET.Element("Domain")


def create_geometry(GeometryType, **attribs):
    element = ET.Element('Geometry')
    element.attrib.update({'GeometryType': GeometryType})
    element.attrib.update(attribs)
    return element


def create_topology(TopologyType, NumberOfElements, **attribs):
    element = ET.Element('Topology')
    element.attrib.update({'TopologyType': TopologyType,
                           'NumberOfElements': NumberOfElements})
    if TopologyType in ['Polyline']:
        if 'NodesPerElement' not in attribs.keys():
            raise Exception('NodesPerElement must be specified')
    element.attrib.update(attribs)
    return element


def create_attribute(Name, **attribs):
    element = ET.Element('Attribute')
    element.attrib.update({'Name': Name})
    element.attrib.update(attribs)
    return element


def create_time(mode='Single', Value=None):
    element = ET.Element('Time')
    if mode == 'Single' and Value:
        element.attrib['Value'] = Value
    return element


def create_grid(Name, GridType, **attribs):
    element = ET.Element('Grid')
    element.attrib.update({'Name': Name,
                           'GridType': GridType,
                           'Section': None})
    element.attrib.update(attribs)
    if element.attrib['GridType'] != 'Subset':
        if 'Section' in element.attrib.keys():
            del element.attrib['Section']
    return element


def create_data_item(value, Dimensions, **attribs):
    element = ET.Element('DataItem')
    element.attrib.update({'ItemType': "Uniform",
                           'Format': "XML",
                           'DataType': "Float",
                           'Precision': "4",
                           'Rank': "1",
                           'Dimensions': Dimensions,
                           'Function': None})
    element.attrib.update(attribs)
    if element.attrib['Function'] is None:
        del element.attrib['Function']
    element.text = value
    return element



import numpy as np
from scipy.interpolate import interp1d


class Solution(np.ndarray):
    """Brief description of 'Solution'"""
    pass


class SolutionContainer(dict):
    """Brief description of 'SolutionContainer'"""
    pass


class SteadyStateSolution(np.ndarray):
    r"""Brief description of 'SteadyStateSolution'"""

    def __new__(cls, x):
        obj = np.asarray(x).view(cls)
        return obj


class UnsteadySolution(Solution):
    """Brief description of 'UnsteadySolution'"""

    def __new__(cls, x, y):
        obj = np.asarray(y).view(cls)
        obj._x = np.asarray(x).view(cls)
        return obj

    def _create_interpolant(self):
        self._interpolant = interp1d(self._x, self, bounds_error=True)

    def interpolate(self, x):
        """
        Interpolates solution at point 'x'.

        Parameters
        ----------
        x : float
            Point at which the solution is to be interpolated

        Returns
        -------
        ndarray
            Solution interpolated at the given point 'x'

        """
        # Cache interpolant to avoid overhead
        if not hasattr(self, "_interpolant"):
            self._create_interpolant()
        return self._interpolant(x)

    __call__ = interpolate


class TransientSolution(UnsteadySolution):

    def interpolate(self, t):
        """
        Interpolates solution at time 't'.

        Parameters
        ----------
        t : float
            Time at which the solution is to be interpolated

        Returns
        -------
        ndarray
            Transient solution interpolated at the given time 't'

        Notes
        -----
        't' must reside inside the 'tspan' used during integration.

        """
        return super().interpolate(x=t)

    @property
    def t(self):
        r"""
        Wrapper to access the generic _x attribute on the super class
        """
        return self._x


class PressureScan(UnsteadySolution):

    def interpolate(self, p):
        s = super().interpolate(x=p).astype(self.dtype)
        return s

    @property
    def p(self):
        r"""
        Wrapper to access the generic _x attribute on the super class
        """
        return self._x

    __call__ = interpolate



import numpy as _np
from openpnm.models.geometry import _geodocs


__all__ = [
    'spheres_and_cylinders',
]


@_geodocs
def spheres_and_cylinders(
    network,
    pore_diameter='pore.diameter',
    throat_diameter='throat.diameter',
    throat_centroid='throat.centroid',
):
    r"""
    Computes the endpoints of throats when pores are spherical.

    The endpoints lie inside the sphere, defined by the lens formed between
    the intersection of the sphere and cylinder.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s
    %(Tcen)s

    Returns
    -------
    endpoints : ndarray
        An array containing the Nt-2-3 coordinates of the end points of each
        throat.
    """
    xyz = network.coords
    cn = network.conns
    L = network['throat.spacing']
    Dt = network[throat_diameter]
    D1 = network[pore_diameter][cn[:, 0]]
    D2 = network[pore_diameter][cn[:, 1]]
    L1 = _np.zeros_like(L)
    L2 = _np.zeros_like(L)
    # Handle the case where Dt > Dp
    mask = Dt > D1
    L1[mask] = 0.5 * D1[mask]
    L1[~mask] = _np.sqrt(D1[~mask]**2 - Dt[~mask]**2) / 2
    mask = Dt > D2
    L2[mask] = 0.5 * D2[mask]
    L2[~mask] = _np.sqrt(D2[~mask]**2 - Dt[~mask]**2) / 2
    # Handle non-colinear pores and throat centroids
    try:
        TC = network[throat_centroid]
        LP1T = _np.linalg.norm(TC - xyz[cn[:, 0]], axis=1) + 1e-15
        LP2T = _np.linalg.norm(TC - xyz[cn[:, 1]], axis=1) + 1e-15
        unit_vec_P1T = (TC - xyz[cn[:, 0]]) / LP1T[:, None]
        unit_vec_P2T = (TC - xyz[cn[:, 1]]) / LP2T[:, None]
    except KeyError:
        unit_vec_P1T = (xyz[cn[:, 1]] - xyz[cn[:, 0]]) / L[:, None]
        unit_vec_P2T = -1 * unit_vec_P1T
    # Find throat endpoints
    EP1 = xyz[cn[:, 0]] + L1[:, None] * unit_vec_P1T
    EP2 = xyz[cn[:, 1]] + L2[:, None] * unit_vec_P2T
    # Handle throats w/ overlapping pores
    L1 = (4 * L**2 + D1**2 - D2**2) / (8 * L)
    L2 = (4 * L**2 + D2**2 - D1**2) / (8 * L)
    h = (2 * _np.sqrt(D1**2 / 4 - L1**2)).real
    overlap = L - 0.5 * (D1 + D2) < 0
    mask = overlap & (Dt < h)
    EP1[mask] = (xyz[cn[:, 0]] + L1[:, None] * unit_vec_P1T)[mask]
    EP2[mask] = (xyz[cn[:, 1]] + L2[:, None] * unit_vec_P2T)[mask]
    return {'head': EP1, 'tail': EP2}



import openpnm.models as mods


water = {
    'pore.contact_angle': {
        'model': mods.misc.constant,
        'value': 110.0,
    },
    'pore.density': {
        'model': mods.phase.density.water_correlation,
    },
    'pore.molar_density': {
        'model': mods.phase.density.mass_to_molar,
    },
    'pore.surface_tension': {
        'model': mods.phase.surface_tension.water_correlation,
    },
    'pore.thermal_conductivity': {
        'model': mods.phase.thermal_conductivity.water_correlation,
    },
    'pore.vapor_pressure': {
        'model': mods.phase.vapor_pressure.liquid_pure_antoine,
    },
    'pore.viscosity': {
        'model': mods.phase.viscosity.water_correlation,
    },
}



import logging

import numpy as np

from openpnm.io import _parse_filename
from openpnm.network import Network

logger = logging.getLogger(__name__)


def network_to_pergeos(network, filename=''):
    # avoid printing truncated array
    np.set_printoptions(threshold=np.inf)

    # Ensure network has PerGeos' expected properties
    if 'pore.EqRadius' not in network.props():
        try:
            network['pore.EqRadius'] = network['pore.diameter']/2
        except KeyError:
            network['pore.EqRadius'] = np.ones([network.Np, ])

    s = ["# Avizo 3D ASCII 3.0\n\n"]
    s.append("define VERTEX " + str(network.Np) + '\n')
    s.append("define EDGE " + str(network.Nt) + '\n')
    s.append("define POINT " + str(2*network.Nt) + '\n\n')
    s.append("Parameters {\n\tContentType \"HxPoreNetworkModel\"\n}\n\n")

    types = {'b': 'int', 'i': 'int', 'f': 'float'}
    typemap = {}
    namemap = {}
    shapemap = {}
    propmap = {}
    i = 1

    NumEdgePoints = 1
    for item in network.keys():
        typemap[item] = types[str(network[item].dtype)[0]]
        ncols = int(network[item].size/network[item].shape[0])
        if ncols > 1:
            shapemap[item] = '[' + str(ncols) + ']'
        else:
            shapemap[item] = ''
        if item.startswith('pore'):
            element = 'pore', 'VERTEX'
        if item.startswith('throat'):
            element = 'throat', 'EDGE'
        n = item.replace(element[0] + '.', '').replace('.', '_').split('_')
        n = ''.join([i[0].upper()+i[1:] for i in n if len(i)])
        namemap[item] = n
        temp = element[1] + " { " + typemap[item] + shapemap[item] + " " \
            + namemap[item] + " } @" + str(i) + '\n'

        if temp.find('EdgeConnectivity') == -1:
            # replaces openpnm tags with the mandatory am file's tags
            if "Conns" in temp:
                temp = temp.replace("Conns", "EdgeConnectivity")
            elif "Coords" in temp:
                temp = temp.replace("Coords", "VertexCoordinates")
            s.append(temp)
        propmap[item] = str(i)
        if "NumEdgePoints" in temp:
            NumEdgePoints = 0
        i += 1

    if NumEdgePoints:
        temp = "EDGE { int NumEdgePoints" + " } @" + str(i) + '\n'
        s.append(temp)
        tempat = "@" + str(i) + '\n'
        i += 1

    # Add POINT data
    s.append("POINT { float[3] EdgePointCoordinates } @" + str(i))
    s.append("\n\n# Data section follows")
    for item in network.keys():
        data = network[item]
        if item != 'throat.EdgeConnectivity':
            s.append('\n\n@' + propmap[item] + '\n')
            if shapemap[item] == '':
                data = np.atleast_2d(data).T
            if typemap[item] == 'float':
                formatter = {'float_kind': lambda x: "%.15E" % x}
            else:
                formatter = None
            if data.dtype == 'bool':
                data = data.astype(int)
            d = np.array2string(data, formatter=formatter)
            s.append(d.replace('[', '').replace(']', '').replace('\n ', '\n'))

    # Add POINT data
    s.append('\n\n@' + str(i) + '\n')
    formatter = {'float_kind': lambda x: "%.15E" % x}

    conns = network['throat.conns']
    d = np.array2string(network['pore.coords'][conns], formatter=formatter)
    for r in (('[', ''), (']', ''), ('\n\n', '\n'), ('\n  ', '\n'),
              ('\n ', '\n')):
        d = d.replace(*r)
    d += '\n'
    s.append(d)

    # Add NumEdgePoints
    if NumEdgePoints:
        s.append('\n\n' + tempat)
        s.append(''.join(['2' + '\n']*network.Nt))

    # Write to file
    if filename == '':
        filename = network.name
    fname = _parse_filename(filename=filename, ext='am')
    with open(fname, 'w') as f:
        f.write(''.join(s))


def network_from_pergeos(filename):
    r"""
    Loads a network from a PerGeos file.

    Notes
    -----
    PerGeos is the format used by the Avizo software. See `here for more
    details <https://cases.pergeos.com/>`_.
    """
    net = {}

    # ---------------------------------------------------------------------
    # Parse the link1 file
    filename = _parse_filename(filename=filename, ext='am')
    with open(filename, mode='r') as f:
        Np = None
        Nt = None
        while (Np is None) or (Nt is None):
            s = f.readline()[:-1].split(' ')
            if s[0] == 'define':
                if s[1] == 'VERTEX':
                    Np = int(s[2])
                if s[1] == 'EDGE':
                    Nt = int(s[2])

        net = {}
        propmap = {}
        typemap = {}
        shapemap = {}
        while True:
            s = f.readline()[:-1].split(' ')
            if s[0] == 'VERTEX':
                dshape = [Np]
                if s[2].endswith(']'):
                    ncols = int(s[2].split('[', 1)[1].split(']')[0])
                    dshape.append(ncols)
                dtype = s[2].split('[')[0]
                temp = np.zeros(dshape, dtype=dtype)
                net['pore.'+s[3]] = temp
                key = int(s[-1].replace('@', ''))
                propmap[key] = 'pore.'+s[3]
                typemap[key] = dtype
                shapemap[key] = dshape
            elif s[0] == 'EDGE':
                dshape = [Nt]
                if s[2].endswith(']'):
                    ncols = int(s[2].split('[', 1)[1].split(']')[0])
                    dshape.append(ncols)
                dtype = s[2].split('[')[0]
                temp = np.zeros(dshape, dtype=dtype)
                net['throat.'+s[3]] = temp
                key = int(s[-1].replace('@', ''))
                propmap[key] = 'throat.'+s[3]
                typemap[key] = dtype
                shapemap[key] = dshape
            elif s[0] == '#':
                break

        s = f.read().split('@')
        for key in propmap.keys():
            if key in s:
                data = s[key].split('\n')[1:]
                data = ' '.join(data)
                arr = np.fromstring(data, dtype=typemap[key], sep=' ')
                arr = np.reshape(arr, shapemap[key])
                net[propmap[key]] = arr
        # End file parsing

    net['pore.coords'] = net['pore.VertexCoordinates']
    net['throat.conns'] = np.sort(net['throat.EdgeConnectivity'], axis=1)

    network = Network()
    network.update(net)

    return network



import logging
from openpnm.algorithms import TransientReactiveTransport, FourierConduction


__all__ = ['TransientFourierConduction']


logger = logging.getLogger(__name__)


class TransientFourierConductionSettings():
    r"""

    Parameters
    ----------
    %(TransientFourierConduction.parameters)s

    """
    quantity = 'pore.temperature'
    conductance = 'throat.thermal_conductance'
    pore_volume = 'pore.volume'


class TransientFourierConduction(TransientReactiveTransport, FourierConduction):
    r"""
    A class to simulate transient thermal diffusion with heat source
    """

    def __init__(self, name='trans_fourier_?', **kwargs):
        super().__init__(name=name, **kwargs)



import logging
import numpy as np
logger = logging.getLogger(__name__)


def network_from_networkx(G):
    r"""
    Creates an OpenPNM Network from a undirected NetworkX graph object

    Parameters
    ----------
    G : networkx.classes.graph.Graph Object
        The NetworkX graph. G should be undirected. The numbering of nodes
        should be numeric (int's), zero-based and should not contain any
        gaps, i.e. ``G.nodes() = [0,1,3,4,5]`` is not allowed and should be
        mapped to ``G.nodes() = [0,1,2,3,4]``.

    Returns
    -------
    network : dict
        An OpenPNM network dictionary

    Notes
    -----
    1. Each node in a NetworkX object (i.e. ``G``) can be assigned properties
    using syntax like ``G.node[n]['diameter'] = 0.5`` where ``n`` is the
    node number.  There is no need to precede the property name with any
    indication that it is pore data such as \'pore\_\'.  OpenPNM will prepend
    \'pore.\' to each property name.

    2. Since \'pore.coords\' is so central to OpenPNM it should be specified
    in the NetworkX object as \'coords\', and the [X, Y, Z] coordinates of
    each node should be a 1x3 list.

    3. Edges in a NetworkX object are accessed using the index numbers of the
    two nodes it connects, such as ``G.adj[2][3]['length'] = 0.1``
    indicating the edge that connects nodes 2 and 3.  There is no need to
    precede the property name with any indication that it is throat data such
    as \'throat\_\'.  OpenPNM will prepend \'throat.\' to each property name.

    4. The \'throat.conns\' property is essential to OpenPNM, but this does NOT
    need to be specified explicitly as a property in NetworkX.  The
    connectivity is embedded into the network representation and is extracted
    by OpenPNM.

    """
    import networkx as nx
    from openpnm.network import Network

    net = {}

    # Ensure G is an undirected networkX graph with numerically numbered
    # nodes for which numbering starts at 0 and does not contain any gaps
    if not isinstance(G, nx.Graph):  # pragma: no cover
        raise Exception('Provided object is not a NetworkX graph.')
    if nx.is_directed(G):  # pragma: no cover
        raise Exception('Provided graph is directed. Convert to undirected graph.')
    if not all(isinstance(n, int) for n in G.nodes()):  # pragma: no cover
        raise Exception('Node numbering is not numeric. Convert to int.')
    if min(G.nodes()) != 0:  # pragma: no cover
        raise Exception('Node numbering does not start at zero.')
    if max(G.nodes()) + 1 != len(G.nodes()):  # pragma: no cover
        raise Exception('Node numbering contains gaps. Map nodes to remove gaps.')

    # Parsing node data
    Np = len(G)
    net.update({'pore.all': np.ones((Np,), dtype=bool)})
    for n, props in G.nodes(data=True):
        for item in props.keys():
            val = props[item]
            dtype = type(val)
            # Remove prepended pore. and pore_ if present
            for b in ['pore.', 'pore_']:
                item = item.replace(b, '')
            # Create arrays for subsequent indexing, if not present already
            if 'pore.'+item not in net.keys():
                if dtype == str:  # handle strings of arbitrary length
                    net['pore.'+item] = np.ndarray((Np,), dtype='object')
                elif dtype is list:
                    dtype = type(val[0])
                    if dtype == str:
                        dtype = 'object'
                    cols = len(val)
                    net['pore.'+item] = np.ndarray((Np, cols), dtype=dtype)
                else:
                    net['pore.'+item] = np.ndarray((Np,), dtype=dtype)
            net['pore.'+item][n] = val

    # Parsing edge data
    # Deal with conns explicitly
    try:
        conns = list(G.edges)   # NetworkX V2
    except Exception:  # pragma: no cover
        conns = G.edges()       # NetworkX V1
    conns.sort()

    # Add conns to Network
    Nt = len(conns)
    net.update({'throat.all': np.ones(Nt, dtype=bool)})
    net.update({'throat.conns': np.array(conns)})

    # Scan through each edge and extract all its properties
    i = 0
    for t in conns:
        props = G[t[0]][t[1]]
        for item in props:
            val = props[item]
            dtype = type(val)
            # Remove prepended throat. and throat_ if present
            for b in ['throat.', 'throat_']:
                item = item.replace(b, '')
            # Create arrays for subsequent indexing, if not present already
            if 'throat.'+item not in net.keys():
                if dtype == str:
                    net['throat.'+item] = np.ndarray((Nt,), dtype='object')
                if dtype is list:
                    dtype = type(val[0])
                    if dtype == str:
                        dtype = 'object'
                    cols = len(val)
                    net['throat.'+item] = np.ndarray((Nt, cols),
                                                     dtype=dtype)
                else:
                    net['throat.'+item] = np.ndarray((Nt,), dtype=dtype)
            net['throat.'+item][i] = val
        i += 1

    network = Network()
    network.update(net)
    return network


def network_to_networkx(network):
    r"""
    Write OpenPNM Network to a NetworkX object.

    Parameters
    ----------
    network : dict
        The OpenPNM Network to be converted to a NetworkX object

    Returns
    -------
    G : undirected graph object
        A NetworkX object with all pore/throat properties attached to it
    """
    import networkx as nx
    from openpnm.network import Network

    # Ensure network is an Network.
    if not isinstance(network, Network):  # pragma: no cover
        raise Exception('Provided network is not an OpenPNM Network.')

    G = nx.Graph()

    # Extracting node list and connectivity matrix from Network
    nodes = map(int, network.Ps)
    conns = network['throat.conns']

    # Explicitly add nodes and connectivity matrix
    G.add_nodes_from(nodes)
    G.add_edges_from(conns)

    # Attach Network properties to G
    for prop in network.props() + network.labels():
        if 'pore.' in prop:
            if len(network[prop].shape) > 1:
                val = {i: list(network[prop][i]) for i in network.Ps}
            else:
                val = {i: network[prop][i] for i in network.Ps}
            nx.set_node_attributes(G, name=prop[5:], values=val)
        if 'throat.' in prop:
            val = {tuple(conn): network[prop][i] for i, conn
                   in enumerate(conns)}
            nx.set_edge_attributes(G, name=prop[7:], values=val)

    return G



import numpy as _np
import openpnm.models.geometry.conduit_lengths as _conduit_lengths
from openpnm.models.geometry import _geodocs


__all__ = [
    "spheres_and_cylinders",
    "circles_and_rectangles",
    "cones_and_cylinders",
    "intersecting_cones",
    "hybrid_cones_and_cylinders",
    "trapezoids_and_rectangles",
    "intersecting_trapezoids",
    "hybrid_trapezoids_and_rectangles",
    "pyramids_and_cuboids",
    "intersecting_pyramids",
    "hybrid_pyramids_and_cuboids",
    "cubes_and_cuboids",
    "squares_and_rectangles",
    "ncylinders_in_series"
]


@_geodocs
def spheres_and_cylinders(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    spheres and throats are cylinders.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------
    size_factors : ndarray
        Array (Nt by 3) containing conduit values for each element
        of the pore-throat-pore conduits. The array is formatted as
        ``[pore1, throat, pore2]``.

    Notes
    -----
    The hydraulic size factor is the geometrical part of the pre-factor in
    Stoke's flow:

    .. math::

        Q = \frac{A^2}{8 \pi \mu L} \Delta P
          = \frac{S_{hydraulic}}{\mu} \Delta P

    Thus :math:`S_{hydraulic}` represents the combined effect of the area
    and length of the *conduit*, which consists of a throat and 1/2 of the
    pores on each end.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.spheres_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    a = 4 / (D1**3 * _np.pi**2)
    b = 2 * D1 * L1 / (D1**2 - 4 * L1**2) + _np.arctanh(2 * L1 / D1)
    F1 = a * b
    a = 4 / (D2**3 * _np.pi**2)
    b = 2 * D2 * L2 / (D2**2 - 4 * L2**2) + _np.arctanh(2 * L2 / D2)
    F2 = a * b
    Ft = Lt / (_np.pi / 4 * Dt**2)**2

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / (2 * _np.pi)

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = 1 / (16 * _np.pi**2 * It * Ft)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def circles_and_rectangles(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    circles and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.circles_and_rectangles(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^3) dx, x = [0, Li]
    F1 = L1 / (D1**2 * _np.sqrt(D1**2 - 4 * L1**2))
    F2 = L2 / (D2**2 * _np.sqrt(D2**2 - 4 * L2**2))
    Ft = Lt / Dt**3

    # S is 1 / (12 * F)
    S1, St, S2 = [1 / (Fi * 12) for Fi in [F1, Ft, F2]]

    return _np.vstack([S1, St, S2]).T


@_geodocs
def cones_and_cylinders(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    truncated cones and throats are cylinders.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.cones_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = 16 / 3 * (L1 * (D1**2 + D1 * Dt + Dt**2) / (D1**3 * Dt**3 * _np.pi**2))
    F2 = 16 / 3 * (L2 * (D2**2 + D2 * Dt + Dt**2) / (D2**3 * Dt**3 * _np.pi**2))
    Ft = Lt / (_np.pi * Dt**2 / 4) ** 2

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / (2 * _np.pi)

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = 1 / (16 * _np.pi**2 * It * Ft)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def intersecting_cones(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes hydraulic size factors of intersecting cones.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.intersecting_cones(
        network=network,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = 16 / 3 * (L1 * (D1**2 + D1 * Dt + Dt**2) / (D1**3 * Dt**3 * _np.pi**2))
    F2 = 16 / 3 * (L2 * (D2**2 + D2 * Dt + Dt**2) / (D2**3 * Dt**3 * _np.pi**2))

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = 1 / (2 * _np.pi)

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)
    St = _np.full(len(Lt), _np.inf)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def hybrid_cones_and_cylinders(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    truncated cones and throats are cylinders.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.hybrid_cones_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = 16 / 3 * (L1 * (D1**2 + D1 * Dt + Dt**2) / (D1**3 * Dt**3 * _np.pi**2))
    F2 = 16 / 3 * (L2 * (D2**2 + D2 * Dt + Dt**2) / (D2**3 * Dt**3 * _np.pi**2))
    Ft = Lt / (_np.pi * Dt**2 / 4) ** 2

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / (2 * _np.pi)

    mask = Lt == 0.0
    if mask.any():
        inv_F_t = _np.zeros(len(Ft))
        inv_F_t[~mask] = 1/Ft[~mask]
        inv_F_t[mask] = _np.inf
    else:
        inv_F_t = 1/Ft
    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = inv_F_t / (16 * _np.pi**2 * It)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def trapezoids_and_rectangles(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    trapezoids and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.cones_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^3) dx, x = [0, Li]
    F1 = L1 / 2 * (D1 + Dt) / (D1 * Dt)**2
    F2 = L2 / 2 * (D2 + Dt) / (D2 * Dt)**2
    Ft = Lt / Dt**3

    # S is 1 / (12 * F)
    S1, St, S2 = [1 / (Fi * 12) for Fi in [F1, Ft, F2]]

    return _np.vstack([S1, St, S2]).T


@_geodocs
def intersecting_trapezoids(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes hydraulic size factors for conduits of intersecting
    trapezoids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.intersecting_trapezoids(
        network=network,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A^3) dx, x = [0, Li]
    F1 = L1 / 2 * (D1 + Dt) / (D1 * Dt)**2
    F2 = L2 / 2 * (D2 + Dt) / (D2 * Dt)**2

    # S is 1 / (12 * F)
    S1 = 1 / (F1 * 12)
    S2 = 1 / (F2 * 12)
    St = _np.full(len(Lt), _np.inf)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def hybrid_trapezoids_and_rectangles(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    trapezoids and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.hybrid_trapezoids_and_rectangles(
        network=network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A^3) dx, x = [0, Li]
    F1 = L1 / 2 * (D1 + Dt) / (D1 * Dt)**2
    F2 = L2 / 2 * (D2 + Dt) / (D2 * Dt)**2
    Ft = Lt / Dt**3

    mask = Lt == 0.0
    if mask.any():
        inv_F_t = _np.zeros(len(Ft))
        inv_F_t[~mask] = 1/Ft[~mask]
        inv_F_t[mask] = _np.inf
    else:
        inv_F_t = 1/Ft

    # S is 1 / (12 * F)
    S1 = 1 / (F1 * 12)
    St = inv_F_t / 12
    S2 = 1 / (F2 * 12)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def pyramids_and_cuboids(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    truncated pyramids and throats are cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.pyramids_and_cuboids(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = 1 / 3 * (L1 * (D1**2 + D1 * Dt + Dt**2) / (D1**3 * Dt**3))
    F2 = 1 / 3 * (L2 * (D2**2 + D2 * Dt + Dt**2) / (D2**3 * Dt**3))
    Ft = Lt / Dt**4

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / 6

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = 1 / (16 * _np.pi**2 * It * Ft)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def intersecting_pyramids(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes hydraulic size factors of intersecting pyramids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.intersecting_pyramids(
        network=network,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = 1 / 3 * (L1 * (D1**2 + D1 * Dt + Dt**2) / (D1**3 * Dt**3))
    F2 = 1 / 3 * (L2 * (D2**2 + D2 * Dt + Dt**2) / (D2**3 * Dt**3))

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = 1 / 6

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)
    St = _np.full(len(Lt), _np.inf)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def hybrid_pyramids_and_cuboids(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    truncated pyramids and throats are cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.hybrid_pyramids_and_cuboids(
        network=network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = 1 / 3 * (L1 * (D1**2 + D1 * Dt + Dt**2) / (D1**3 * Dt**3))
    F2 = 1 / 3 * (L2 * (D2**2 + D2 * Dt + Dt**2) / (D2**3 * Dt**3))
    Ft = Lt / Dt**4

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / 6

    mask = Lt == 0.0
    if mask.any():
        inv_F_t = _np.zeros(len(Ft))
        inv_F_t[~mask] = 1/Ft[~mask]
        inv_F_t[mask] = _np.inf
    else:
        inv_F_t = 1/Ft

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = inv_F_t / (16 * _np.pi**2 * It)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def cubes_and_cuboids(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    pore_aspect=[1, 1, 1],
    throat_aspect=[1, 1, 1],
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are cubes
    and throats are cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s
    pore_aspect : list
        Aspect ratio of the pores
    throat_aspect : list
        Aspect ratio of the throats

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.cubes_and_cuboids(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^2) dx, x = [0, Li]
    F1 = L1 / D1**4
    F2 = L2 / D2**4
    Ft = Lt / Dt**4

    # I is the integral of (y^2 + z^2) dA, divided by A^2
    I1 = I2 = It = 1 / 6

    # S is 1 / (16 * pi^2 * I * F)
    S1 = 1 / (16 * _np.pi**2 * I1 * F1)
    St = 1 / (16 * _np.pi**2 * It * Ft)
    S2 = 1 / (16 * _np.pi**2 * I2 * F2)

    return _np.vstack([S1, St, S2]).T


@_geodocs
def squares_and_rectangles(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    pore_aspect=[1, 1],
    throat_aspect=[1, 1],
):
    r"""
    Computes hydraulic size factors for conduits assuming pores are
    squares and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s
    pore_aspect : list
        Aspect ratio of the pores
    throat_aspect : list
        Aspect ratio of the throats

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    L1, Lt, L2 = _conduit_lengths.squares_and_rectangles(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T

    # Fi is the integral of (1/A^3) dx, x = [0, Li]
    F1 = L1 / D1**3
    F2 = L2 / D2**3
    Ft = Lt / Dt**3

    # S is 1 / (12 * F)
    S1, St, S2 = [1 / (Fi * 12) for Fi in [F1, Ft, F2]]

    return _np.vstack([S1, St, S2]).T


@_geodocs
def ncylinders_in_series(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    n=5,
):
    r"""
    Computes hydraulic size factors for conduits of spheres and cylinders
    with the spheres approximated as N cylinders in series.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s
    n : int
        Number of cylindrical divisions for each pore

    Returns
    -------

    Notes
    -----

    """
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split('.', 1)[-1]).T
    # Ensure throats are never bigger than connected pores
    Dt = _np.minimum(Dt, 0.99 * _np.minimum(D1, D2))
    L1, Lt, L2 = _conduit_lengths.spheres_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    ).T
    dL1 = _np.linspace(0, L1, num=n)
    dL2 = _np.linspace(0, L2, num=n)
    r1 = D1 / 2 * _np.sin(_np.arccos(dL1 / (D1 / 2)))
    r2 = D2 / 2 * _np.sin(_np.arccos(dL2 / (D2 / 2)))

    gtemp = (_np.pi * r1 ** 4 / (8 * L1 / n)).T
    F1 = 1 / _np.sum(1 / gtemp, axis=1)
    gtemp = (_np.pi * r2 ** 4 / (8 * L2 / n)).T
    F2 = 1 / _np.sum(1 / gtemp, axis=1)
    Ft = (_np.pi * (Dt / 2) ** 4 / (8 * Lt)).T

    return _np.vstack([F1, Ft, F2]).T



import numpy as np
from openpnm.models.phase import _phasedocs


__all__ = [
    "water_correlation",
    "liquid_pure_antoine",
    "liquid_pure_lk",
]


@_phasedocs
def water_correlation(phase, T='pore.temperature', salinity='pore.salinity'):
    r"""
    Calculates vapor pressure of pure water or seawater given by [1] based on
    Raoult's law. The pure water vapor pressure is given by [2]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(salinity)s

    Returns
    -------


    Notes
    -----
     T must be in K, and S in g of salt per kg of phase, or ppt (parts per
        thousand)
    VALIDITY: 273 < T < 473 K; 0 < S < 240 g/kg;
    ACCURACY: 0.5 percent

    References
    ----------
    [1] Sharqawy M. H., Lienhard J. H., and Zubair, S. M., Desalination and
    Water Treatment, 2010.
    [2] ASHRAE handbook: Fundamentals, ASHRAE; 2005.

    """
    T = phase[T]
    if salinity in phase.keys():
        S = phase[salinity]
    else:
        S = 0
    a1 = -5.8002206E+03
    a2 = 1.3914993E+00
    a3 = -4.8640239E-02
    a4 = 4.1764768E-05
    a5 = -1.4452093E-08
    a6 = 6.5459673E+00
    Pv_w = np.exp((a1/T) + a2 + a3*T + a4*T**2 + a5*T**3 + a6*np.log(T))
    Pv_sw = Pv_w/(1+0.57357*(S/(1000-S)))
    value = Pv_sw
    return value


@_phasedocs
def liquid_pure_lk(
    phase,
    T='pore.temperature',
    Tc='param.critical_temperature',
    Pc='param.critical_pressure',
    omega='param.acentric_factor',
):
    r"""
    Calculate the vapor pressure of a pure liquid using the correlation in [1]

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(Tc)s
    %(Pc)s
    %(omega)s

    Returns
    -------

    References
    ----------
    [1] Lee and Kesler
    """
    T = phase[T]
    Tc = phase[Tc]
    Tr = T/Tc
    Pc = phase[Pc]
    omega = phase[omega]
    f0 = 5.92714 - 6.09648/Tr - 1.28862*np.log(Tr) + 0.169347*(Tr**6)
    f1 = 15.2518 - 15.6875/Tr - 13.4721*np.log(Tr) + 0.43577*(Tr**6)
    Pr = np.exp(f0 + omega*f1)
    Pvap = Pr*Pc
    return Pvap


@_phasedocs
def liquid_pure_antoine(
    phase,
    T='pore.temperature',
):
    r"""
    Calculates the vapor pressure of a pure liquid using Antoine's equation

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(Tc)s

    Returns
    -------

    References
    ----------
    [1] RPP

    Notes
    -----
    Coefficients are looked up from the ``chemicals`` package. If values for
    "extended" version are not found then the normal 3-coefficient values are
    sought.

    """
    # either antoine or extended antoine using constants from RPP
    CAS = phase.params['CAS']
    T = phase[T]
    from chemicals.vapor_pressure import Psat_data_AntoinePoling
    coeffs = Psat_data_AntoinePoling.loc[CAS]
    _, A, B, C, Tmin, Tmax = coeffs
    Pvap = 10**(A - B/(T + C))
    return Pvap



import numpy as np

from openpnm.models.geometry import _geodocs

__all__ = [
    "spheres_and_cylinders",
    "circles_and_rectangles",
    "cones_and_cylinders",
    "intersecting_cones",
    "hybrid_cones_and_cylinders",
    "trapezoids_and_rectangles",
    "hybrid_trapezoids_and_rectangles",
    "intersecting_trapezoids",
    "pyramids_and_cuboids",
    "intersecting_pyramids",
    "hybrid_pyramids_and_cuboids",
    "cubes_and_cuboids",
    "squares_and_rectangles",
]


@_geodocs
def spheres_and_cylinders(
    network,
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
):
    r"""
    Calculates conduit lengths in the network assuming pores are spheres
    and throats are cylinders.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------
    lengths : ndarray
        Array (Nt by 3) containing conduit values for each element
        of the pore-throat-pore conduits. The array is formatted as
        ``[pore1, throat, pore2]``.

    """
    L_ctc = _get_L_ctc(network)
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split(".", 1)[-1]).T

    # Handle the case where Dt > Dp
    if (Dt > D1).any() or (Dt > D2).any():
        _raise_incompatible_data()

    # If spheres do not overlap:
    L1 = np.sqrt(D1**2 - Dt**2) / 2
    L2 = np.sqrt(D2**2 - Dt**2) / 2
    Lt = L_ctc - (L1 + L2)

    if np.any(Lt < 0):  # Find pores that touch/overlap
        # Find diameter of overlap between spheres
        d = L_ctc
        R1 = D1/2
        R2 = D2/2
        # Check distance to the intersection
        L1_int = (d**2 - R2**2 + R1**2) / (2*d)
        if np.any(L1_int < 0) or np.any(L1_int > L_ctc):
            raise Exception('The pores overlap too much')
        D_int = 2/(2*d) * np.sqrt(4*d**2 * R1**2 - (d**2 - R2**2 + R1**2)**2)
        mask = D_int > Dt
        if np.any(mask):
            d = d[mask]
            R1 = R1[mask]
            R2 = R2[mask]
            L1[mask] = (d**2 - R2**2 + R1**2) / (2*d)
            L2[mask] = L_ctc[mask] - L1[mask]
            Lt[mask] = 1e-15

    return np.vstack((L1, Lt, L2)).T


@_geodocs
def circles_and_rectangles(
    network, pore_diameter="pore.diameter", throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are circles
    and throats are rectangles.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    return spheres_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter,
    )


@_geodocs
def cones_and_cylinders(
    network, pore_diameter="pore.diameter", throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are cones
    and throats are cylinders.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    L_ctc = _get_L_ctc(network)
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split(".", 1)[-1]).T

    L1 = D1 / 2
    L2 = D2 / 2

    # Handle throats w/ overlapping pores
    _L1 = (4 * L_ctc**2 + D1**2 - D2**2) / (8 * L_ctc)
    mask = L_ctc - 0.5 * (D1 + D2) < 0
    L1[mask] = _L1[mask]
    L2[mask] = (L_ctc - L1)[mask]
    Lt = np.maximum(L_ctc - (L1 + L2), 1e-15)
    return np.vstack((L1, Lt, L2)).T


@_geodocs
def intersecting_cones(network, pore_coords="pore.coords", throat_coords="throat.coords"):
    r"""
    Calculates conduit lengths in the network assuming pores are cones
    that intersect. Therefore, the throat is the cross sectional plane where
    two pores meet and has negligible/zero volume.

    A conduit is defined as ( 1/2 pore - 1/2 pore ).

    Parameters
    ----------
    %(network)s
    %(Pcoords)s
    %(Tcoords)s

    Returns
    -------

    """
    P12 = network["throat.conns"]
    p_coords = network[pore_coords]
    t_coords = network[throat_coords]
    L1 = np.sqrt(np.sum(((p_coords[P12[:, 0]] - t_coords)) ** 2, axis=1))
    L2 = np.sqrt(np.sum(((p_coords[P12[:, 1]] - t_coords)) ** 2, axis=1))
    Lt = np.zeros(len(network.Ts))
    return np.vstack((L1, Lt, L2)).T


@_geodocs
def hybrid_cones_and_cylinders(
    network, pore_diameter="pore.diameter", throat_coords="throat.coords"
):
    r"""
    Calculates conduit lengths in the network assuming pores are cones
    and throats are cylinders.

    A conduit is defined as ( 1/2 pore - full throat - 1/2 pore ).

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    """
    L_ctc = _get_L_ctc(network)
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split(".", 1)[-1]).T

    L1 = D1 / 2
    L2 = D2 / 2
    Lt = np.maximum(L_ctc - (L1 + L2), 1e-15)
    # Handle intersecting pores
    XYp = network.coords[network.conns]
    XYt = network[throat_coords]
    _L1, _L2 = np.linalg.norm(XYp - XYt[:, None], axis=2).T
    mask1 = _L1 < L1
    mask2 = _L2 < L2
    mask = np.logical_or(mask1, mask2)
    if mask.any():
        L1[mask] = _L1[mask]
        L2[mask] = _L2[mask]
        Lt[mask] = 0.0
    return np.vstack((L1, Lt, L2)).T


@_geodocs
def trapezoids_and_rectangles(
    network, pore_diameter="pore.diameter", throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are
    trapezoids and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    return cones_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter,
    )


@_geodocs
def intersecting_trapezoids(
    network, pore_coords="pore.coords", throat_coords="throat.coords"
):
    r"""
    Calculates conduit lengths in the network assuming pores are
    intersecting trapezoids.

    Parameters
    ----------
    %(network)s
    %(Pcoords)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    return intersecting_cones(
        network=network,
        pore_coords=pore_coords,
        throat_coords=throat_coords,
    )


@_geodocs
def hybrid_trapezoids_and_rectangles(
    network, pore_diameter="pore.diameter", throat_coords="throat.coords"
):
    r"""
    Calculates conduit lengths in the network assuming pores are
    trapezoids and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    return hybrid_cones_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords,
    )


@_geodocs
def pyramids_and_cuboids(
    network, pore_diameter="pore.diameter", throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are truncated
    pyramids and throats are cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    return cones_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter,
    )


@_geodocs
def intersecting_pyramids(
    network, pore_coords="pore.coords", throat_coords="throat.coords"
):
    r"""
    Calculates conduit lengths in the network assuming pores are
    intersecting pyramids.

    Parameters
    ----------
    %(network)s
    %(Pcoords)s
    %(Tcoords)s

    Returns
    -------

    """
    return intersecting_cones(
        network=network,
        pore_coords=pore_coords,
        throat_coords=throat_coords,
    )


@_geodocs
def hybrid_pyramids_and_cuboids(
    network, pore_diameter="pore.diameter", throat_coords="throat.coords"
):
    r"""
    Calculates conduit lengths in the network assuming pores are truncated
    pyramids that intersect. Therefore, the throat is the cross sectional plane where
    two pores meet and has negligible/zero volume.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    """
    return hybrid_cones_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords,
    )


@_geodocs
def cubes_and_cuboids(
    network, pore_diameter="pore.diameter", throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are cubes
    and throats are cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    L_ctc = _get_L_ctc(network)
    D1, Dt, D2 = network.get_conduit_data(pore_diameter.split(".", 1)[-1]).T

    L1 = D1 / 2
    L2 = D2 / 2
    Lt = L_ctc - (L1 + L2)

    # Handle overlapping pores
    mask = (Lt < 0) & (L1 > L2)
    L2[mask] = (L_ctc - L1)[mask]
    mask = (Lt < 0) & (L2 > L1)
    L1[mask] = (L_ctc - L2)[mask]
    Lt = np.maximum(Lt, 1e-15)
    return np.vstack((L1, Lt, L2)).T


@_geodocs
def squares_and_rectangles(
    network, pore_diameter="pore.diameter", throat_diameter="throat.diameter"
):
    r"""
    Calculates conduit lengths in the network assuming pores are squares
    and throats are rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    Notes
    -----
    This model should only be used for true 2D networks, i.e. with planar
    symmetry.

    """
    return cubes_and_cuboids(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter,
    )


# Dealing with errors and exceptions
def _raise_incompatible_data():
    msg = (
        "'spheres_and_cylinders' can only be applied when throat diameter is"
        " smaller than that of adjacent pores."
    )
    raise Exception(msg)


def _get_L_ctc(network):
    """Returns throat spacing if it exists, otherwise calculates it."""
    try:
        L_ctc = network["throat.spacing"]
    except KeyError:
        P12 = network["throat.conns"]
        C1 = network["pore.coords"][P12[:, 0]]
        C2 = network["pore.coords"][P12[:, 1]]
        L_ctc = np.linalg.norm(C1 - C2, axis=1)
    return L_ctc



import numpy as np
from openpnm.models.phase import _phasedocs


__all__ = [
    "liquid_mixture_tc",
    "gas_mixture_ce",
    "gas_mixture_fesg",
    # "gas_mixture_wilke_fuller",
]


@_phasedocs
def liquid_mixture_tc(
    phase,
    T='pore.temperature',
    mu='pore.viscosity',
    Vms_at_Tb='param.molar_volume_Tb.*',
    sigmas_at_Tb='param.surface_tension_Tb.*',
):
    r"""
    Uses Tyn-Calus model to estimate diffusion coefficient in a dilute liquid
    solution of A in B from first principles at conditions of interest

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(mu)s
    Vms_at_Tb : str or list of scalars
        Molar volumes of each component at its boiling temperature (m3/mol).
        Can either be a string to a parameter on each component of a list
        of scalar values which are used directly
    sigmas_at_Tb : float
        Surface tension of component A at boiling temperature (N/m)

    Returns
    -------

    """
    T = phase[T]
    mu = phase[mu]
    sigma_A, sigma_B = phase.get_comp_vals(sigmas_at_Tb).values()
    VA, VB = phase.get_comp_vals(Vms_at_Tb).values()
    A = 8.93e-8*(VB*1e6)**0.267/(VA*1e6)**0.433*T
    B = (sigma_B/sigma_A)**0.15/(mu*1e3)
    value = A*B*1e-4
    return value


@_phasedocs
def gas_mixture_ce(
    phase,
    T='pore.temperature',
    P='pore.pressure',
    Tcs='param.critical_temperature.*',
    Pcs='param.critical_pressure.*',
    omegas='param.acentric_factor.*',
    MWs='param.molecular_weight.*',
    epsilons='param.LJ_energy.*',
    sigmas='param.LJ_diameter.*',
):
    r"""
    Calculate gas phase diffusion coefficient using Chapman-Enskog equation.

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(P)s
    %(Tcs)s
    %(Pcs)s
    %(omegas)s
    %(MWs)s
    %(epsilons)s
    %(sigmas)s

    Returns
    -------

    """
    # Fetch values from components
    T = phase[T]
    P = phase[P]
    try:
        MA, MB = phase.get_comp_vals(MWs).values()
    except ValueError:
        raise Exception('This function only works on binary mixtures')
    MWAB = 2/(1/MA + 1/MB)
    omega = phase.get_comp_vals(omegas).values()
    Tc = phase.get_comp_vals(Tcs).values()
    Pc = phase.get_comp_vals(Pcs).values()
    k = 1.380649e-23  # Boltzmann constant
    try:
        eA, eB = phase.get_comp_vals(epsilons).values()
    except Exception:
        # Use correlation of Tee, Gotoh, & Stewart
        eA, eB = (0.7915 + 0.1693*omega)*Tc*k
    eAB = (eA*eB)**0.5
    # Compute collision integral using Neufeld's correlation (RPP Eq.(11-3.6))
    A, B, C, D, E, F, G, H = (1.06036, 0.15610, 0.19300, 0.47635, 1.03587,
                              1.52996, 1.76474, 3.89411)
    Tstar = k*T/eAB
    Omega = Omega = A/(Tstar**B) + C/np.exp(D*Tstar) + E/np.exp(F*Tstar) \
        + G/np.exp(H*Tstar)
    # Now apply RPP Eq.(11-3.2)
    try:
        sA, sB = phase.get_comp_vals(sigmas).values()
    except Exception:
        # Use correlation of Tee, Gotoh, & Stewart
        sA, sB = (2.3551 - 0.08847*omega)*(Tc/Pc)**(1/3)
    sAB = (sA + sB)/2
    DAB = 0.00266 * (T**1.5) / (P/101325 * MWAB**0.5 * sAB**2 * Omega) * 1e-4
    return DAB


@_phasedocs
def gas_mixture_fesg(
    phase,
    T='pore.temperature',
    P='pore.pressure',
    MWs='param.molecular_weight.*',
    Vdms='param.molar_diffusion_volume.*',
):
    r"""
    Estimates the diffusion coefficient of both species in a binary gas
    mixture using the Fuller et al correlation [1]_, [2]_, [3]_.

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(P)s
    %(MWs)s
    %(Vdms)s

    Returns
    -------
    Dij : dict[ndarray]
        The dict contains one array for each component, containing the
        diffusion coefficient of that component at each location.

    References
    ----------
    .. [1] Fuller, E. N., and J. C. Giddings: J. Gas Chromatogr., 3: 222 (1965).
    .. [2] Fuller, E. N., P. D. Schettler, and J. C. Giddings: Ind. Eng. Chem.,
       58(5): 18 (1966).
    .. [3] Fuller, E. N., K. Ensley, and J. C. Giddings: J. Phys. Chem.,
       73: 3679 (1969).

    """
    T = phase[T]
    P = phase[P]
    # The following is to accomodate proper mixtures AND normal Phase objects
    try:
        MA, MB = phase.get_comp_vals(MWs).values()
    except AttributeError:
        MA, MB = phase[MWs]
    try:
        vA, vB = phase.get_comp_vals(Vdms).values()
    except AttributeError:
        vA, vB = phase[Vdms]
    MAB = 2/(1/MA + 1/MB)
    P = P/101325
    DAB = 0.00143*T**1.75/(P*(MAB**0.5)*(vA**(1./3) + vB**(1./3))**2)*1e-4
    return DAB


def gas_mixture_fw(
    phase,
    T='pore.temperature',
    P='pore.pressure',
    MWs='param.molecular_weight.*',
    Vdms='pore.molar_diffusion_volume.*',
):  # pragma: no cover
    r"""
    Estimates the diffusion coeffient of each species in a gas mixture

    Uses the Fuller equation to estimate binary diffusivity between pairs,
    then uses the correction of Fairbanks and Wilke [1]_ to account for the
    composition of the gas mixture.

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(P)s
    %(MWs)s
    %(Vdms)s

    Returns
    -------
    Dij : dict[ndarray]
        The dict contains one array for each component, containing the
        diffusion coefficient of that component at each location.

    References
    ----------
    .. [1] Fairbanks DF and CR Wilke, Diffusion Coefficients in Multicomponent
       Gas Mixtures. Industrial & Engineering Chemistry, 42(3), p471–475 (1950).
       `DOI: 10.1021/ie50483a022 <http://doi.org/10.1021/ie50483a022>`_

    """
    raise Exception('This function is not ready yet')
    # comps = list(phase.components.values())
    # values = {}
    # for i in range(len(comps)):
    #     A = comps[i]
    #     denom = 0.0
    #     for j in range(len(comps)):
    #         if i != j:
    #             B = comps[j]
    #             D = gas_mixture_fesg(phase=phase,
    #                                  molecular_weight=MW_pair,
    #                                  molar_diffusion_volume=Vdm_pair,
    #                                  temperature=T,
    #                                  pressure=P)
    #             yB = phase['pore.mole_fraction.' + B.name]
    #             denom += yB/D
    #     yA = phase['pore.mole_fraction.' + A.name]
    #     values[A.name] = (1 - yA)/denom
    # return values



try:
    from matplotlib._docstring import Substitution
except ModuleNotFoundError:
    from matplotlib.docstring import Substitution



__all__ = [
    '_doctxt',
]


_doctxt = Substitution(
    phase=
    r"""phase : OpenPNM Phase object
            The phase object to which this model is associated (i.e. attached).
            This controls the length of the calculated array(s), and also
            provides access to other necessary properties.""",
    network=
    r"""network : OpenPNM Network object
            The network object to which this model is associated (i.e.
            attached). This controls the length of the calculated array(s),
            and also provides access to other necessary properties.""",
    dict_blurb=
    r"""Name of the dictionary key on the target object pointing to the
    array containing values of """,
    return_arr=
    r"""values : ndarray
            A numpy ndarray containing the computed values of """,
)



r"""
Algorithms
==========

This module contains various algorithms for performing transport simulations


"""

from ._algorithm import *
from ._transport import *

from ._reactive_transport import *
from ._transient_reactive_transport import *

from ._stokes_flow import *

from ._fickian_diffusion import *
from ._transient_fickian_diffusion import *

from ._advection_diffusion import *
from ._transient_advection_diffusion import *

from ._fourier_conduction import *
from ._transient_fourier_conduction import *

from ._ohmic_conduction import *

from ._drainage import *
from ._invasion_percolation import *



import numpy as np
import scipy.spatial as sptl
from openpnm._skgraph.generators import tools
from openpnm._skgraph.tools import tri_to_am, isoutside
from openpnm._skgraph.operations import trim_nodes


def delaunay(
    points,
    shape=[1, 1, 1],
    reflect=False,
    f=1,
    trim=True,
    node_prefix='node',
    edge_prefix='edge',
):
    r"""
    Generate a network based on Delaunay triangulation of random points

    Parameters
    ----------
    points : array_like or int
        Can either be an N-by-3 array of point coordinates which will be used,
        or a scalar value indicating the number of points to generate
    shape : array_like
        Indicates the size and shape of the domain
    reflect : boolean, optional (default = ``False``)
        If ``True`` then points are reflected across each face of the domain
        prior to performing the tessellation. These reflected points are
        automatically trimmed.  Enabling this behavior prevents long-range
        connections between surface pores.
    f : float
        The fraction of points which should be reflected.  The default is 1 which
        reflects all the points in the domain, but this can lead to a lot of
        unnecessary points, so setting to 0.1 or 0.2 helps speed, but risks that
        the tessellation may not have smooth faces if not enough points are
        reflected.
    trim : boolean, optional (default = ``True``)
        If ``True`` then any points laying outside the domain are removed. This is
        mostly only useful if ``reflect=True``.

    Returns
    -------
    network : dict
        A dictionary containing 'node.coords' and 'edge.conns'
    tri : Delaunay tessellation object
        The Delaunay tessellation object produced by ``scipy.spatial.Delaunay``
    """
    points = tools.parse_points(points=points, shape=shape, reflect=reflect, f=f)
    mask = ~np.all(points == 0, axis=0)
    tri = sptl.Delaunay(points=points[:, mask])
    coo = tri_to_am(tri)
    d = {}
    d[node_prefix+'.coords'] = points
    d[edge_prefix+'.conns'] = np.vstack((coo.row, coo.col)).T
    if trim:
        trim = isoutside(d, shape=shape)
        d = trim_nodes(network=d, inds=np.where(trim)[0])
    return d, tri



import numpy as np
from openpnm._skgraph import tools


__all__ = [
    'join',
]


def join(g1, g2, L_max=0.99):
    r"""
    Joins two networks together topologically including new connections

    Parameters
    ----------
    g1 : dictionary
        A dictionary containing 'node.coords' and 'edge.conns'.
    g2 : dictionary
        A dictionary containing 'node.coords' and 'edge.conns'
    L_max : float
        The distance between nodes below which they are joined

    Returns
    -------
    network : dict
        A new dictionary containing coords and conns from both ``net1`` and
        ``net2``, plus new connections between the two input networks.

    Notes
    -----
    The returned network will use the same node and edge prefixes as ``g1``

    """
    node_prefix_1 = tools.get_node_prefix(g1)
    node_prefix_2 = tools.get_node_prefix(g2)
    if node_prefix_1 != node_prefix_2:
        for item in g2.keys():
            _, prop = item.split('.', 1)
            g2[node_prefix_1 + '.' + item] = g2.pop(item)
    node_prefix = node_prefix_1
    edge_prefix_1 = tools.get_edge_prefix(g1)
    edge_prefix_2 = tools.get_edge_prefix(g2)
    if edge_prefix_1 != edge_prefix_2:
        for item in g2.keys():
            _, prop = item.split('.', 1)
            g2[edge_prefix_1 + '.' + item] = g2.pop(item)
    edge_prefix = edge_prefix_1

    # Perform neighbor query
    from scipy.spatial import KDTree
    t1 = KDTree(g1[node_prefix+'.coords'])
    t2 = KDTree(g2[node_prefix+'.coords'])
    pairs = t1.query_ball_tree(t2, r=L_max)
    # Combine existing network data
    net3 = {}
    Np1 = g1[node_prefix+'.coords'].shape[0]
    Np2 = g2[node_prefix+'.coords'].shape[0]
    Nt1 = g1[edge_prefix+'.conns'].shape[0]
    Nt2 = g2[edge_prefix+'.conns'].shape[0]
    net3[node_prefix+'.coords'] = \
        np.vstack((g1.pop(node_prefix+'.coords'),
                   g2.pop(node_prefix+'.coords')))
    net3[edge_prefix+'.conns'] = \
        np.vstack((g1.pop(edge_prefix+'.conns'),
                   g2.pop(edge_prefix+'.conns') + Np1))
    # Convert kdtree result into new connections
    nnz = sum([len(row) for row in pairs])
    conns = np.zeros((nnz, 2), dtype=int)
    i = 0
    for j, row in enumerate(pairs):
        for col in row:
            conns[i, :] = j, col + Np1
            i += 1
    # Add new connections to network
    net3[edge_prefix+'.conns'] = \
        np.vstack((net3.pop(edge_prefix+'.conns'), conns))

    # Finally, expand any other data arrays on given networks
    items = list(g1.keys()) + list(g2.keys())
    N = {node_prefix: [Np1, Np2], edge_prefix: [Nt1, Nt2]}
    for item in items:
        prefix, attr = item.split('.', 1)
        if prefix == node_prefix:
            temp1 = g1.pop(item, None)
            temp2 = g2.pop(item, None)
        if prefix == edge_prefix:
            temp1 = g1.pop(item, None)
            temp2 = g2.pop(item, None)
        if temp1 is None:
            blank = False if temp2.dtype == bool else np.nan
            temp1 = np.zeros(N[prefix][0], dtype=type(blank))*blank
        elif temp2 is None:
            blank = False if temp1.dtype == bool else np.nan
            temp2 = np.zeros(N[prefix][1], dtype=type(blank))*blank
        net3[item] = np.concatenate((temp1, temp2), axis=0)
    return net3



import heapq as hq
import logging
from collections import namedtuple

import numpy as np
from numba import jit, njit
from tqdm.auto import tqdm

from openpnm._skgraph.queries import qupc_initialize, qupc_reduce, qupc_update
from openpnm._skgraph.simulations import bond_percolation, site_percolation
from openpnm.algorithms import Algorithm
from openpnm.utils import Docorator

__all__ = [
    'InvasionPercolation',
]


logger = logging.getLogger(__name__)
docstr = Docorator()


@docstr.get_sections(base='IPSettings',
                     sections=['Parameters', 'Other Parameters'])
@docstr.dedent
class IPSettings:
    r"""

    Parameters
    ----------
    %(AlgorithmSettings.parameters)s
    pore_volume : str
        The dictionary key for the pore volume array
    throat_volume : str
        The dictionary key for the throat volume array
    entry_pressure : str
        The dictionary key for the throat capillary pressure

    """
    phase = ''
    pore_volume = 'pore.volume'
    throat_volume = 'throat.volume'
    entry_pressure = 'throat.entry_pressure'


class InvasionPercolation(Algorithm):
    r"""
    A classic invasion percolation algorithm optimized for speed with numba

    Parameters
    ----------
    network : Network
        The Network upon which the invasion will occur

    Notes
    -----
    This algorithm uses a `binary heap <https://en.wikipedia.org/wiki/Binary_heap>`_
    to store a list of all accessible throats, sorted according to entry
    pressure.  This means that item [0] in the heap is the most easily invaded
    throat that is currently accessible by the invading fluid, so looking up
    which throat to invade next is computationally trivial. In order to keep
    the list sorted, adding new throats to the list takes more time; however,
    the heap data structure is very efficient at this.

    """

    def __init__(self, phase, name='ip_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(IPSettings())
        self.settings['phase'] = phase.name
        self['pore.bc.inlet'] = False
        self['pore.bc.outlet'] = False
        self.reset()

    def reset(self):
        self['pore.invasion_sequence'] = -1
        self['throat.invasion_sequence'] = -1
        self['pore.trapped'] = False
        self['throat.trapped'] = False
        # self['pore.residual'] = False
        # self['throat.residual'] = False

    def _set_residual(self, pores=None, throats=None, mode='add'):  # pragma: no cover
        raise NotImplementedError("The ability to add residual nwp is not ready yet")
        if mode == 'add':
            if pores is not None:
                self['pore.residual'][pores] = True
            if throats is not None:
                self['throat.residual'][throats] = True
        elif mode == 'drop':
            if pores is not None:
                self['pore.residual'][pores] = False
            if throats is not None:
                self['throat.residual'][throats] = False
        elif mode == 'clear':
            if pores is not None:
                self['pore.residual'] = False
            if throats is not None:
                self['throat.residual'] = False
        elif mode == 'overwrite':
            if pores is not None:
                self['pore.residual'] = False
                self['pore.residual'][pores] = True
            if throats is not None:
                self['throat.residual'] = False
                self['throat.residual'][throats] = True

    def set_inlet_BC(self, pores=None, mode='add'):
        r"""
        Specifies which pores are treated as inlets for the invading phase

        Parameters
        ----------
        pores : ndarray
            The indices of the pores from which the invading fluid invasion
            should start
        mode : str or list of str, optional
            Controls how the boundary conditions are applied. Options are:

            ============ =====================================================
            mode         meaning
            ============ =====================================================
            'add'        (default) Adds the supplied boundary conditions to
                         the given locations. Raises an exception if values
                         of any type already exist in the given locations.
            'overwrite'  Adds supplied boundary conditions to the given
                         locations, including overwriting conditions of the
                         given type or any other type that may be present in
                         the given locations.
            'remove'     Removes boundary conditions of the specified type
                         from the specified locations. If ``bctype`` is not
                         specified then *all* types are removed. If no
                         locations are given then values are remvoed from
                         *all* locations.
            ============ =====================================================

            If a list of strings is provided, then each mode in the list is
            handled in order, so that ``['remove', 'add']`` will give the same
            results add ``'overwrite'``.

        """
        self.set_BC(pores=pores, bcvalues=True, bctype='inlet', mode=mode)
        self.reset()
        self['pore.invasion_sequence'][self['pore.bc.inlet']] = 0

    def set_outlet_BC(self, pores=None, mode='add'):
        r"""
        Specifies which pores are treated as outlets for the defending phase

        This must be specified if trapping is to be considered.

        Parameters
        ----------
        pores : ndarray
            The indices of the pores from which the defending fluid exits the
            domain
        mode : str or list of str, optional
            Controls how the boundary conditions are applied. Options are:

            ============ =====================================================
            mode         meaning
            ============ =====================================================
            'add'        (default) Adds the supplied boundary conditions to
                         the given locations. Raises an exception if values
                         of any type already exist in the given locations.
            'overwrite'  Adds supplied boundary conditions to the given
                         locations, including overwriting conditions of the
                         given type or any other type that may be present in
                         the given locations.
            'remove'     Removes boundary conditions of the specified type
                         from the specified locations. If ``bctype`` is not
                         specified then *all* types are removed. If no
                         locations are given then values are remvoed from
                         *all* locations.
            ============ =====================================================

            If a list of strings is provided, then each mode in the list is
            handled in order, so that ``['remove', 'add']`` will give the same
            results add ``'overwrite'``.

        """
        self.set_BC(pores=pores, bcvalues=True, bctype='outlet', mode=mode)

    def run(self):
        r"""
        Performs the algorithm for the given number of steps

        """
        # Setup arrays and info
        # TODO: This should be called conditionally so that it doesn't
        # overwrite existing data when doing a few steps at a time
        self._run_setup()
        n_steps = np.inf

        # Create incidence matrix for use in _run_accelerated which is jit
        im = self.network.create_incidence_matrix(fmt='csr')

        # Perform initial analysis on input pores
        Ts = self.project.network.find_neighbor_throats(pores=self['pore.bc.inlet'])
        t_start = self['throat.order'][Ts]
        t_inv, p_inv, p_inv_t = \
            _run_accelerated(
                t_start=t_start,
                t_sorted=self['throat.sorted'],
                t_order=self['throat.order'],
                t_inv=self['throat.invasion_sequence'],
                p_inv=self['pore.invasion_sequence'],
                p_inv_t=np.zeros_like(self['pore.invasion_sequence']),
                conns=self.project.network['throat.conns'],
                idx=im.indices,
                indptr=im.indptr,
                n_steps=n_steps)

        # Transfer results onto algorithm object
        self['throat.invasion_sequence'] = t_inv
        self['pore.invasion_sequence'] = p_inv
        self['throat.invasion_pressure'] = self['throat.entry_pressure']
        self['pore.invasion_pressure'] = self['throat.entry_pressure'][p_inv_t]
        # Set invasion pressure of inlets to 0
        self['pore.invasion_pressure'][self['pore.invasion_sequence'] == 0] = 0.0
        # Set invasion sequence and pressure of any residual pores/throats to 0
        # self['throat.invasion_sequence'][self['throat.residual']] = 0
        # self['pore.invasion_sequence'][self['pore.residual']] = 0

    def _run_setup(self):
        self['pore.invasion_sequence'][self['pore.bc.inlet']] = 0
        # self['pore.invasion_sequence'][self['pore.residual']] = 0
        # self['throat.invasion_sequence'][self['throat.residual']] = 0
        # Set throats between inlets as trapped
        Ts = self.network.find_neighbor_throats(self['pore.bc.inlet'], mode='xnor')
        self['throat.trapped'][Ts] = True
        # Get throat capillary pressures from phase and update
        phase = self.project[self.settings['phase']]
        self['throat.entry_pressure'] = phase[self.settings['entry_pressure']]
        # self['throat.entry_pressure'][self['throat.residual']] = 0.0
        # Generated indices into t_entry giving a sorted list
        self['throat.sorted'] = np.argsort(self['throat.entry_pressure'], axis=0)
        self['throat.order'] = 0
        self['throat.order'][self['throat.sorted']] = np.arange(0, self.Nt)

    def pc_curve(self):
        r"""
        Get the percolation data as the non-wetting phase saturation vs the
        capillary pressure.

        """
        net = self.project.network
        pvols = net[self.settings['pore_volume']]
        tvols = net[self.settings['throat_volume']]
        tot_vol = np.sum(pvols) + np.sum(tvols)
        # Normalize
        pvols /= tot_vol
        tvols /= tot_vol
        # Remove trapped volume
        pmask = np.isfinite(self['pore.invasion_sequence'])
        tmask = np.isfinite(self['throat.invasion_sequence'])
        pvols = pvols[pmask]
        tvols = tvols[tmask]
        pseq = self['pore.invasion_sequence'][pmask]
        tseq = self['throat.invasion_sequence'][tmask]
        pPc = self['pore.invasion_pressure'][pmask]
        tPc = self['throat.invasion_pressure'][tmask]
        vols = np.concatenate((pvols, tvols))
        seqs = np.concatenate((pseq, tseq))
        Pcs = np.concatenate((pPc, tPc))
        data = np.rec.fromarrays([seqs, vols, Pcs],
                                 formats=['i', 'f', 'f'],
                                 names=['seq', 'vol', 'Pc'])
        data.sort(axis=0, order='seq')
        pc_curve = namedtuple('pc_curve', ('pc', 'snwp'))
        data = pc_curve(data.Pc, np.cumsum(data.vol))
        return data

    def apply_trapping(self):
        r"""
        Adjusts the invasion sequence of pores and throats that are trapped.

        This method uses the reverse invasion percolation procedure outlined
        by Masson [1].

        Returns
        -------
        This function does not return anything. It adjusts the
        ``'pore.invasion_sequence'`` and ``'throat.invasion_sequence'`` arrays
        on the object by setting trapped pores/throats to ``ninf``. It also puts
        ``True`` values into the ``'pore.trapped'`` and ``'throat.trapped'``
        arrays.

        Notes
        -----
        Outlet pores must be specified (using ``set_outlet_BC`` or putting
        ``True`` values in ``alg['pore.bc.outlet']``) or else an exception is
        raised.

        References
        ----------
        [1] Masson, Y. https://doi.org/10.1016/j.cageo.2016.02.003

        """
        outlets = np.where(self['pore.bc.outlet'])[0]
        am = self.network.create_adjacency_matrix(fmt='csr')
        inv_seq = self['pore.invasion_sequence']
        self['pore.trapped'] = _find_trapped_pores(inv_seq, am.indices,
                                                   am.indptr, outlets)
        # Update invasion sequence
        self['pore.invasion_sequence'][self['pore.trapped']] = -1
        # Find which throats are trapped, including throats which were invaded
        # after both of it's pores were invaded (hence have a unique invasion
        # sequence number).
        pmask = self['pore.invasion_sequence'][self.network.conns]
        tmask = np.stack((self['throat.invasion_sequence'],
                          self['throat.invasion_sequence'])).T
        hits = ~np.any(pmask == tmask, axis=1)
        self['throat.trapped'] = hits
        self['throat.invasion_sequence'][hits] = -1
        # Make some adjustments
        Pmask = self['pore.invasion_sequence'] < 0
        Tmask = self['throat.invasion_sequence'] < 0
        self['pore.invasion_sequence'] = \
            self['pore.invasion_sequence'].astype(float)
        self['pore.invasion_sequence'][Pmask] = np.inf
        self['throat.invasion_sequence'] = \
            self['throat.invasion_sequence'].astype(float)
        self['throat.invasion_sequence'][Tmask] = np.inf

    def _apply_trapping_slow(self, step_size=1, mode='mixed'):  # pragma: no cover
        # TODO: Make sure this function actually works and keep it for debugging
        N = self['throat.invasion_sequence'].max()
        pseq = self['pore.invasion_sequence']
        tseq = self['throat.invasion_sequence']
        msg = 'Evaluating trapping'
        for i in tqdm(range(0, int(N), step_size), msg):
            if mode == 'bond':
                i += 1
                s, b = bond_percolation(conns=self.network.conns,
                                        occupied_bonds=tseq > i)
            elif mode == 'site':
                s, b = site_percolation(conns=self.network.conns,
                                        occupied_sites=pseq > i)
            clusters = np.unique(s[self['pore.bc.outlet']])
            self['pore.trapped'] += np.isin(s, clusters, invert=True)*(pseq > i)
            self['throat.trapped'] += np.isin(b, clusters, invert=True)*(tseq > i)
        # Set trapped pores/throats to uninvaded and adjust invasion sequence
        self['pore.invasion_sequence'][self['pore.trapped']] = -1
        self['throat.invasion_sequence'][self['throat.trapped']] = -1
        # Set any residual pores within trapped clusters back to untrapped
        # self['pore.trapped'][self['pore.residual']] = False
        # self['throat.trapped'][self['throat.residual']] = False


@jit(forceobj=True)
def _find_trapped_pores(inv_seq, indices, indptr, outlets):
    Np = len(inv_seq)
    sorted_seq = np.vstack((inv_seq.astype(np.int_), np.arange(Np, dtype=np.int_))).T
    sorted_seq = sorted_seq[sorted_seq[:, 0].argsort()][::-1]
    cluster = -np.ones(Np, dtype=np.int_)
    trapped_pores = np.zeros(Np, dtype=bool)
    trapped_clusters = np.zeros(Np, dtype=bool)
    # cluster_map = qupc_initialize(Np)
    cluster_map = np.arange(Np, dtype=np.int_)
    next_cluster_num = 0
    i = -1
    for step, pore in sorted_seq:
        i += 1
        step, pore = sorted_seq[i, :]
        n = indices[indptr[pore]:indptr[pore+1]]
        nc = cluster_map[cluster[n]][inv_seq[n] > step]
        nc_uniq = np.unique(nc)
        if nc.size == 0:
            # Found an isolated pore, start a new cluster
            cluster[pore] = next_cluster_num
            # If pore is an outlet then note cluster as no longer trapped
            if pore in outlets:
                trapped_clusters[next_cluster_num] = False
            else:  # Otherwise note this cluster as being a trapped cluster
                trapped_clusters[next_cluster_num] = True
                # Note this pore as trapped as well
                trapped_pores[pore] = True
            # Increment cluster number for next time
            next_cluster_num += 1
        elif nc_uniq.size == 1:
            c = nc_uniq[0]
            # Neighbors have one unique cluster number, so assign it to current pore
            cluster[pore] = c
            # If pore is an outlet then note cluster as no longer trapped
            if pore in outlets:
                trapped_clusters[c] = False
                # Also set all joined clusters to not trapped
                cluster_map = qupc_reduce(cluster_map)
                hits = np.where(cluster_map == cluster_map[c])[0]
                trapped_clusters[hits] = False
            # If this cluster number is part of a trapped cluster then
            # mark pore as trapped
            if trapped_clusters[c]:
                trapped_pores[pore] = True
        elif nc_uniq.size > 1:
            cluster[pore] = min(nc_uniq)
            # Merge all clusters into a single cluster
            for c in nc:
                qupc_update(cluster_map, c, min(nc_uniq))
            cluster_map = qupc_reduce(cluster_map)
            # If all neighboring clusters are trapped, then set current pore to
            # trapped as well
            if np.all(trapped_clusters[nc]):
                trapped_pores[pore] = True
            else:  # Otherwise set all neighbor clusters to untrapped!
                trapped_clusters[nc] = False
    return trapped_pores


@njit
def _run_accelerated(t_start, t_sorted, t_order, t_inv, p_inv, p_inv_t,
                     conns, idx, indptr, n_steps):  # pragma: no cover
    r"""
    Numba-jitted run method for InvasionPercolation class.

    Notes
    -----
    ``idx`` and ``indptr`` are properties are the network's incidence
    matrix, and are used to quickly find neighbor throats.

    Numba doesn't like foreign data types (i.e. Network), and so
    ``find_neighbor_throats`` method cannot be called in a jitted method.

    Nested wrapper is for performance issues (reduced OpenPNM import)
    time due to local numba import

    """
    # TODO: The following line is supposed to be numba's new list, but the
    # heap does not work with this
    # queue = List(t_start)
    queue = list(t_start)
    hq.heapify(queue)
    count = 1
    while count < (n_steps + 1):
        # Find throat at the top of the queue
        t = hq.heappop(queue)
        # Extract actual throat number
        t_next = t_sorted[t]
        t_inv[t_next] = count
        # If throat is duplicated
        while len(queue) > 0 and queue[0] == t:
            _ = hq.heappop(queue)
        # Find pores connected to newly invaded throat from am in coo format
        Ps = conns[t_next]
        # Remove already invaded pores from Ps
        Ps = Ps[p_inv[Ps] < 0]
        # If either of the neighboring pores are uninvaded (-1), set it to
        # invaded and add its neighboring throats to the queue
        if len(Ps) > 0:
            p_inv[Ps] = count
            p_inv_t[Ps] = t_next
            for i in Ps:
                # Get neighboring throat numbers from im in csr format
                Ts = idx[indptr[i]:indptr[i+1]]
                # Keep only throats which are uninvaded
                Ts = Ts[t_inv[Ts] < 0]
            for i in Ts:  # Add throat to the queue
                hq.heappush(queue, t_order[i])
        count += 1
        if len(queue) == 0:
            break
    return t_inv, p_inv, p_inv_t


# %%
if __name__ == '__main__':
    import matplotlib.pyplot as plt

    import openpnm as op

    np.random.seed(0)
    Nx, Ny, Nz = 25, 25, 1
    pn = op.network.Cubic(shape=[Nx, Ny, Nz], spacing=1e-4)
    pn.add_model_collection(op.models.collections.geometry.spheres_and_cylinders)
    pn.regenerate_models()
    pn['pore.volume@left'] = 0.0
    # op.topotools.trim(pn, pores=[380, 395])

    water = op.phase.Water(network=pn, name='h2o')
    water.add_model_collection(op.models.collections.physics.standard)
    water.regenerate_models()

    ip = InvasionPercolation(network=pn, phase=water)
    ip.set_inlet_BC(pn.pores('left'))
    ip.run()
    ip.set_outlet_BC(pn.pores('right'))
    # ip.apply_trapping()

    # %%
    if 0:
        pseq = np.copy(ip['pore.invasion_sequence'])
        tseq = np.copy(ip['throat.invasion_sequence'])
        ax = op.topotools.plot_connections(pn, tseq >= 0, linestyle='--', c='b',
                                           linewidth=3)
        op.topotools.plot_coordinates(pn, pseq >= 0, c='r', marker='x',
                                      markersize=100, ax=ax)
        op.topotools.plot_coordinates(pn, pseq < 0, c='c', marker='o',
                                      markersize=100, ax=ax)

    # %%
    if 1:
        drn = op.algorithms.Drainage(network=pn, phase=water)
        drn.set_inlet_BC(pn.pores('left'))
        pressures = np.unique(ip['pore.invasion_pressure'])
        # pressures = np.logspace(np.log10(0.1e3), np.log10(2e4), 100)
        drn.run(pressures=pressures)
        drn.set_outlet_BC(pn.pores('right'))
        # drn.apply_trapping()

        fig, ax = plt.subplots(1, 1)
        ax.step(*ip.pc_curve(), 'b', where='post')
        ax.step(*drn.pc_curve(), 'r--', where='post')
        ax.set_ylim([0, 1])

    # %%
    if 0:
        pseq = ip['pore.invasion_sequence']
        pseq[pseq == -1] = pseq.max() + 1
        tseq = ip['throat.invasion_sequence']
        tseq[tseq == -1] = tseq.max() + 1
        for j, i in enumerate(tqdm(np.unique(tseq)[:-1])):
            ax = op.topotools.plot_connections(pn, tseq <= i, linestyle='--',
                                               c='r', linewidth=3)
            op.topotools.plot_coordinates(pn, pseq <= i, c='b', marker='x',
                                          markersize=100, ax=ax)
            op.topotools.plot_coordinates(pn, ip['pore.trapped'], c='c',
                                          marker='o', markersize=100, ax=ax)
            plt.savefig(f"{str(j).zfill(3)}.png")
            plt.close()



r"""
Collection of functions for manipulating network topology
=========================================================

This module contains a selection of functions that deal specifically with
network topology.

"""

from ._topotools import *
from ._perctools import *
from ._graphtools import *



from scipy.sparse import csr_matrix, csc_matrix
from scipy.sparse.linalg import spsolve, cg
from openpnm.solvers import DirectSolver, IterativeSolver

__all__ = ['ScipySpsolve', 'ScipyCG']


class ScipySpsolve(DirectSolver):
    """Solves a linear system using ``scipy.sparse.linalg.spsolve``."""

    def solve(self, A, b, **kwargs):
        """Solves the given linear system of equations Ax=b."""
        if not isinstance(A, (csr_matrix, csc_matrix)):
            A = A.tocsr()
        return (spsolve(A, b), 0)


class ScipyCG(IterativeSolver):
    """Solves a linear system using ``scipy.sparse.linalg.cg``."""

    def solve(self, A, b, **kwargs):
        """Solves the given linear system of equations Ax=b."""
        if not isinstance(A, (csr_matrix, csc_matrix)):
            A = A.tocsr()
        atol = self._get_atol(b)
        try:
            return cg(A, b, tol=self.tol, atol=atol, **kwargs)
        except TypeError:
            return cg(A, b, rtol=self.tol, atol=atol, **kwargs)



from numpy.linalg import norm

__all__ = ['BaseSolver', 'DirectSolver', 'IterativeSolver']


class BaseSolver:
    """Base class for all solvers."""
    def __init__(self):
        ...

    def solve(self):
        """Solves the given linear system of equations Ax=b."""
        raise NotImplementedError


class DirectSolver(BaseSolver):
    """Base class for all direct solvers."""
    ...


class IterativeSolver(BaseSolver):
    """Base class for iterative solvers."""
    def __init__(self, tol=1e-8, maxiter=1000):
        self.tol = tol
        self.maxiter = maxiter
        self.atol = None        # needs to be evaluated later
        self.rtol = None        # needs to be evaluated later

    def _get_atol(self, b):
        r"""
        Returns the absolute tolerance ``atol`` that corresponds to the
        the given tolerance ``tol``.

        Notes
        -----
        ``atol`` is calculated to satisfy the following stopping criterion:
            ``norm(A*x-b)`` <= ``atol``

        """
        return norm(b) * self.tol

    def _get_rtol(self, A, b, x0):
        r"""
        Returns the relative tolerance ``rtol`` that corresponds to the
        the given tolerance ``tol``.

        Notes
        -----
        ``rtol`` is defined based on the following formula:
            ``rtol = residual(@x_final) / residual(@x0)``

        """
        res0 = self._get_residual(A, b, x0)
        atol = self._get_atol(b)
        rtol = atol / res0
        return rtol

    def _get_residual(self, A, b, x):
        r"""
        Calculates the residual based on the given ``x`` using:
            ``res = norm(A*x - b)``
        """
        return norm(A * x - b)



try:
    from matplotlib._docstring import Substitution
except ModuleNotFoundError:
    from matplotlib.docstring import Substitution


__all__ = [
    '_phasedocs',
]


_phasedocs = Substitution(
    phase=
    r"""phase : OpenPNM Phase object
            The phase object to which this model is associated (i.e. attached).
            This controls the length of the calculated array(s), and also
            provides access to other necessary properties.""",
    T=
    r"""T : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            temperature values, in units of [K].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    P=
    r"""P : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            pressure values, in units of [Pa].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    MWs=
    r"""MWs : str (dict key) or dict of numeric values for each species

    """,
    MW=
    r"""MW : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            molecular weight values, in units of [g/mol].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    Tcs=
    r"""Tcs : str (dict key) or scalar

    """,
    Tc=
    r"""Tc : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            critical temperature values, in units of [K].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    Pcs=
    r"""Pcs : str (dict key) or scalar

    """,
    Pc=
    r"""Pc : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            critical pressure values, in units of [Pa].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    Tb=
    r"""Tb : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            boiling temperature values, in units of [K].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    Vcs=
    r"""Vcs : str (dict key) or scalar

    """,
    Vc=
    r"""Vc : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            critical volume values, in units of [m^3].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    omegas=
    r"""omegas : str (dict key) or scalar

    """,
    omega=
    r"""omega : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            accentric factor values, which is dimensionless.
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    epsilons=
    r"""epsilons : str (dict key) or scalar

    """,
    epsilon=
    r"""epsilon : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            Lennard-Jones epsilon values, in units of [K].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    sigmas=
    r"""sigmas : str (dict key) or scalar

    """,
    sigma=
    r"""sigma : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            Lennard-Jones molecular diameter values, in units of [A].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    Cpg=
    r"""Cpg : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            gas phase heat capacity values, in units of [J/mol].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    Cps=
    r"""Cps : str (dict key) or scalar

    """,
    ks=
    r"""ks : str (dict key) or numeric
    """,
    Vdms=
    r"""

    """,
    Vm=
    r"""Vm : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            molar volume, in units of [m3/mol].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    n_V=
    r"""n_V : str (dict key) or scalar

    """,
    rhos=
    r"""rhos : str
            The dictionary key containing the density in kg/m3""",
    rho=
    r"""rho : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            mass density values, in units of [kg/m^3].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    mus=
    r"""mus : str

    """,
    mu=
    r"""mu : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            kinematic viscosity values, in units of [Pa.s].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    SInote=
    r"""Since numpy arrays don't natively support units, OpenPNM cannot enforce
    units on values and arrays, however the use of SI is assumed.""",
    salinity=
    r"""salinity : str
            The dictionary key containing the salinity values in g/kg""",
    conc=
    r"""conc : str (dict key) or numeric
            Name of the dictionary key on ``phase`` containing the ndarray of
            concentration values, in units of [mol/L].
            If a numerical value is passed (i.e. a scalar or ndarray)
            it gets used directly.""",
    zs=
    r"""T : str (dict key)
            Name of the dictionary key on ``phase`` containing the ndarray of
            mole fraction values.""",
)



import re

import numpy as np
from pandas import read_csv

from openpnm.io import _parse_filename
from openpnm.io._pandas import network_to_pandas, project_to_pandas
from openpnm.utils import Project


def project_to_csv(project, filename=''):
    r"""
    Save all the pore and throat data on the Network and Phase objects to a CSV
    file

    Parameters
    ----------
    project : list
        An openpnm ``project`` object
    filename : str or path object
        The name of the file to store the data

    """
    df = project_to_pandas(project=project, join=True, delim='.')
    if filename == '':
        filename = project.name
    fname = _parse_filename(filename=filename, ext='csv')
    df.to_csv(fname, index=False)


def network_to_csv(network, filename=''):
    """Exports a network to a CSV file."""
    proj = Project()
    proj.append(network)
    df = network_to_pandas(network=network, join=True, delim='.')
    if filename == '':
        filename = network.name
    fname = _parse_filename(filename=filename, ext='csv')
    df.to_csv(fname, index=False)


def network_from_csv(filename):
    """Loads a network from a CSV file."""
    from openpnm.network import Network
    fname = _parse_filename(filename=filename, ext='csv')

    a = read_csv(filepath_or_buffer=fname,
                 sep=',',
                 skipinitialspace=True,
                 header=0,
                 low_memory=False)

    # First parse through all the items and re-merge columns
    dct = {}
    keys = sorted(list(a.keys()))
    for item in keys:
        m = re.search(r'\[.\]', item)  # The dot '.' is a wildcard
        if m:  # m is None if pattern not found, otherwise merge cols
            pname = re.split(r'\[.\]', item)[0]  # Get base propname
            # Find all other keys with same base propname
            merge_keys = [k for k in a.keys() if k.startswith(pname)]
            # Rerieve and remove arrays with same base propname
            merge_cols = [a.pop(k) for k in merge_keys]
            # Merge arrays into multi-column array and store in DataFrame
            dct[pname] = np.vstack(merge_cols).T
            # Remove key from list of keys
            for k in keys:
                if k.startswith(pname):
                    keys.pop(keys.index(k))
        else:
            dct[item] = np.array(a.pop(item))

    # Now scan through 'pore.coords' and 'throat.conns' to get Np and Nt,
    # then remove the nans
    try:
        Np = np.where(np.isnan(dct['pore.coords'][:, 0]))[0][0]
    except IndexError:
        Np = dct['pore.coords'][:, 0].shape[0]
    try:
        Nt = np.where(np.isnan(dct['throat.conns'][:, 0]))[0][0]
    except IndexError:
        Nt = dct['throat.conns'][:, 0].shape[0]
    for k, v in dct.items():
        if k.startswith('pore.'):
            dct[k] = v[:Np, ...]
        if k.startswith('throat.'):
            dct[k] = v[:Nt, ...]

    # This is a fix for pandas loading in all data types as dtype('O')
    # if it encounters mixed types in the file. For whatever reason
    # the conversion to type 'float64' works without problem, however for
    # boolean values, the dtype 'O' persists. That can lead to difficult
    # to debug errors later on. Therefore here this fix:
    for k, v in dct.items():
        if type(v[0]) is bool:
            dct[k] = v.astype(bool)

    network = Network()
    network.update(dct)
    return network



r"""
Collection of matrix solvers for steady state simulations
=========================================================

The ``solvers`` module contains wrapper classes for sparse matrix solvers.

"""

from ._base import *
from ._scipy import *
from ._pardiso import *
from ._petsc import *
from ._pyamg import *
